
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the linalg routines
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "global.h"

static complex v[25];
static complex w[25];
static spinor *ppk[10];


int main(int argc,char *argv[])
{
   int my_rank,count,i,j,n;
   int vol,icom;
   float r;
   complex z,mz;
   double wt1,wt2,wdt;
   spinor *pk,*pl;
   FILE *log=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      log=freopen("time1.log","w",stdout);
      
      printf("\n");
      printf("Timing of the linalg routines\n");
      printf("------------------------------\n\n");   
      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   vol=VOLUME;
   icom=1;
   
   start_ranlux(0,12345);
   geometry();
   alloc_s(10);

   for (i=0;i<10;i++)
      random_s(vol,ps[i][0],1.0f);

   n=4000000/vol;
   if (n<1)
      n=1;

   pk=ps[0][0];
   pl=ps[1][0];

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
      z=spinor_prod(vol,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function spinor_prod:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(96.0/wdt),sizeof(spinor)/3);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      r=spinor_prod_re(vol,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function spinor_prod_re:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(48.0/wdt),sizeof(spinor)/3);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      r=norm_square(vol,icom,pk);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function norm_square:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(48.0/wdt),sizeof(spinor)/3);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      normalize(vol,icom,pk);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function normalize:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(72.0/wdt),sizeof(spinor)/3);
   }
   
   n/=4;
   if (n<1)
      n=1;

   mz.re=-z.re;
   mz.im=-z.im;

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
   {
      mulc_spinor_add(vol,pk,pl,z);
      mulc_spinor_add(vol,pk,pl,mz);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(2*n*vol);

   if (my_rank==0)
   {   
      printf("Function mulc_spinor_add:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(96.0/wdt),sizeof(spinor)/3);
   }
   
   random_s(vol,ps[0][0],1.0f);
   random_s(vol,ps[1][0],1.0f);
   normalize(vol,icom,pk);
   normalize(vol,icom,pl);
   project(vol,icom,pl,pk);
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
      project(vol,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function project:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(192.0/wdt),sizeof(spinor)/3);
   }
   
   for (i=0;i<5;i++)
   {
      random_s(vol,ps[i][0],1.0f);
      ppk[i]=ps[i][0];

      for (j=0;j<i;j++)
         project(vol,icom,ppk[i],ppk[j]);

      normalize(vol,icom,ppk[i]);
   }

   for (i=5;i<10;i++)
   {
      random_s(vol,ps[i][0],1.0f);
      ppk[i]=ps[i][0];

      for (j=5;j<i;j++)
         project(vol,icom,ppk[i],ppk[j]);

      normalize(vol,icom,ppk[i]);
   }   
      
   for (i=0;i<5;i++)
   {
      for (j=0;j<5;j++)
      {
         z=spinor_prod(vol,icom,ppk[i],ppk[j+5]);

         v[5*i+j].re= z.re;
         v[5*i+j].im= z.im;
         w[5*j+i].re= z.re;
         w[5*j+i].im=-z.im;         
      }
   }

   n/=100;
   if (n<1)
      n=1;
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
   {
      rotate(vol,5,ppk,v);
      rotate(vol,5,ppk,w);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   error_chk();   
   wdt=1.0e6*(wt2-wt1)/(double)(50*n*vol);
   
   if (my_rank==0)
   {   
      printf("Function rotate:\n");
      printf("Time per lattice point & (eigenvalue)^2: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(96.0/wdt),sizeof(spinor)/3);
      fclose(log);
   }

   MPI_Finalize();   
   exit(0);
}
