
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Various direct checks on the cabmar program [right-invariance, 
* invariance of the action under over-relaxation moves and deviation
* from SU(3)]
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"

#define N 1000


static void random_staples(su3 *v)
{
   int i;
   float *r;

   r=(float*)(v);
   gauss(r,18);

   for (i=0;i<18;i++)
      r[i]*=3.0f;
}


static void renormalize(su3 *u)
{
   float sq;
   complex z;

   sq=
      (*u).c11.re*(*u).c11.re+(*u).c11.im*(*u).c11.im+
      (*u).c12.re*(*u).c12.re+(*u).c12.im*(*u).c12.im+
      (*u).c13.re*(*u).c13.re+(*u).c13.im*(*u).c13.im;

   sq=1.5f-0.5f*sq;
   (*u).c11.re*=sq;
   (*u).c11.im*=sq;
   (*u).c12.re*=sq;
   (*u).c12.im*=sq;
   (*u).c13.re*=sq;
   (*u).c13.im*=sq;   

   z.re=
      (*u).c11.re*(*u).c21.re+(*u).c11.im*(*u).c21.im+
      (*u).c12.re*(*u).c22.re+(*u).c12.im*(*u).c22.im+
      (*u).c13.re*(*u).c23.re+(*u).c13.im*(*u).c23.im;
   z.im=
      (*u).c11.re*(*u).c21.im-(*u).c11.im*(*u).c21.re+
      (*u).c12.re*(*u).c22.im-(*u).c12.im*(*u).c22.re+
      (*u).c13.re*(*u).c23.im-(*u).c13.im*(*u).c23.re;

   (*u).c21.re-=(z.re*(*u).c11.re-z.im*(*u).c11.im);
   (*u).c21.im-=(z.re*(*u).c11.im+z.im*(*u).c11.re);
   (*u).c22.re-=(z.re*(*u).c12.re-z.im*(*u).c12.im);
   (*u).c22.im-=(z.re*(*u).c12.im+z.im*(*u).c12.re);
   (*u).c23.re-=(z.re*(*u).c13.re-z.im*(*u).c13.im);
   (*u).c23.im-=(z.re*(*u).c13.im+z.im*(*u).c13.re);    

   sq=
      (*u).c21.re*(*u).c21.re+(*u).c21.im*(*u).c21.im+
      (*u).c22.re*(*u).c22.re+(*u).c22.im*(*u).c22.im+
      (*u).c23.re*(*u).c23.re+(*u).c23.im*(*u).c23.im;

   sq=1.5f-0.5f*sq;
   (*u).c21.re*=sq;
   (*u).c21.im*=sq;
   (*u).c22.re*=sq;
   (*u).c22.im*=sq;
   (*u).c23.re*=sq;
   (*u).c23.im*=sq;

   (*u).c31.re=
      ((*u).c12.re*(*u).c23.re-(*u).c12.im*(*u).c23.im)-
      ((*u).c22.re*(*u).c13.re-(*u).c22.im*(*u).c13.im);
   (*u).c31.im=
      ((*u).c22.re*(*u).c13.im+(*u).c22.im*(*u).c13.re)-
      ((*u).c12.re*(*u).c23.im+(*u).c12.im*(*u).c23.re);
   (*u).c32.re=
      ((*u).c13.re*(*u).c21.re-(*u).c13.im*(*u).c21.im)-
      ((*u).c23.re*(*u).c11.re-(*u).c23.im*(*u).c11.im);
   (*u).c32.im=
      ((*u).c23.re*(*u).c11.im+(*u).c23.im*(*u).c11.re)-
      ((*u).c13.re*(*u).c21.im+(*u).c13.im*(*u).c21.re);
   (*u).c33.re=
      ((*u).c11.re*(*u).c22.re-(*u).c11.im*(*u).c22.im)-
      ((*u).c21.re*(*u).c12.re-(*u).c21.im*(*u).c12.im);
   (*u).c33.im=
      ((*u).c21.re*(*u).c12.im+(*u).c21.im*(*u).c12.re)-
      ((*u).c11.re*(*u).c22.im+(*u).c11.im*(*u).c22.re);   
}
      

int main(int argc,char *argv[])
{
   int my_rank,count,i;
   float beta,r1,r2;
   double d,dmax,dmax_all,s_old,s_new;
   su3 u,v,up,vp,w,z;
   FILE *log=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      log=freopen("check3.log","w",stdout);
   
      printf("\n");
      printf("Check of the program 'cabmar' (invariance, normalization)\n");
      printf("---------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);       
   }   
   
   beta=12.0;   
   start_ranlux(0,12345);

   dmax=0.0;

   for (count=0;count<N;count++)
   {
      random_su3(&u);
      random_staples(&v);
      random_su3(&w);
      _su3_times_su3(up,u,w);
      _su3_times_su3(vp,v,w);
      
      cabmar(beta,&u,&v,1);
      cabmar(beta,&up,&vp,1);

      _su3_times_su3(vp,u,w);      

      d=0.0;

      for (i=0;i<18;i++)
      {
         r1=*((float*)(&up)+i);
         r2=*((float*)(&vp)+i);         
         d+=((double)(r1)-(double)(r2))*((double)(r1)-(double)(r2));
      }

      d=sqrt(d);
      if (d>dmax)
         dmax=d;      
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {    
      printf("Absolute deviation ||u-u'||_F (right-invariance): ");
      printf("%.2e\n\n",dmax_all);
   }
   
   dmax=0.0;

   for (count=0;count<N;count++)
   {
      random_su3(&u);
      random_staples(&v);
      _su3_dagger(w,v);      

      _su3_times_su3(z,u,w);
      s_old=(double)(z.c11.re+z.c22.re+z.c33.re);
      
      cabmar(beta,&u,&v,1);

      _su3_times_su3(z,u,w);
      s_new=(double)(z.c11.re+z.c22.re+z.c33.re);

      d=fabs(s_old-s_new);
      if (d>dmax)
         dmax=d;
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {   
      printf("Variation of the action (over-relaxation): ");
      printf("%.2e\n\n",dmax_all);
   }
   
   dmax=0.0;
   
   for (count=0;count<N;count++)
   {
      random_su3(&u);
      random_staples(&v);
      cabmar(beta,&u,&v,0);

      v=u;
      renormalize(&u);

      d=0.0;

      for (i=0;i<18;i++)
      {
         r1=*((float*)(&u)+i);
         r2=*((float*)(&v)+i);         
         d+=((double)(r1)-(double)(r2))*((double)(r1)-(double)(r2));
      }

      d=sqrt(d);
      if (d>dmax)
         dmax=d;
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {    
      printf("Absolute deviation ||u-u'||_F from SU(3) (heatbath): ");
      printf("%.2e\n\n",dmax_all);    
   }
   
   dmax=0.0;
   
   for (count=0;count<N;count++)
   {
      random_su3(&u);
      random_staples(&v);
      cabmar(beta,&u,&v,1);

      v=u;
      renormalize(&u);

      d=0.0;

      for (i=0;i<18;i++)
      {
         r1=*((float*)(&u)+i);
         r2=*((float*)(&v)+i);         
         d+=((double)(r1)-(double)(r2))*((double)(r1)-(double)(r2));
      }

      d=sqrt(d);
      if (d>dmax)
         dmax=d;
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {    
      printf("Absolute deviation ||u-u'||_F from SU(3) (over-relaxation): ");
      printf("%.2e\n\n",dmax_all);   
      fclose(log);
   }   

   MPI_Finalize();  
   exit(0);
}
