
/*******************************************************************************
*
* File check8.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program that translates the single-precision gauge field
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "random.h"
#include "update.h"
#include "global.h"

static su3 *utot0,*utot1,*ub;


static int cmp_su3(su3 *u,su3 *v)
{
   int k;
   float *ru,*rv;

   ru=(float*)(u);
   rv=(float*)(v);   

   for (k=0;k<18;k++)
   {
      if (ru[k]!=rv[k])
         return 1;
   }

   return 0;
}


static void get_field(void)
{
   int my_rank,ix,tag,n;
   MPI_Status stat;
   
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      for (ix=0;ix<(4*VOLUME);ix++)
         utot1[ix]=ub[ix];
   }
   
   for (n=1;n<NPROC;n++)
   {
      tag=mpi_tag();
      
      if (my_rank==n)
         MPI_Send((float*)(ub),4*18*VOLUME,MPI_FLOAT,
                  0,tag,MPI_COMM_WORLD);

      if (my_rank==0)
         MPI_Recv((float*)(utot1+n*4*VOLUME),4*18*VOLUME,MPI_FLOAT,
                  n,tag,MPI_COMM_WORLD,&stat);
   }
}


static void cpy_field(void)
{
   int my_rank,ix;
   
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      for (ix=0;ix<(4*VOLUME*NPROC);ix++)
         utot0[ix]=utot1[ix];
   }

}


static int cmp_field(int *s)
{
   int my_rank,itest,is;
   int ip,iq,ix,iy,mu;
   int l0,l1,l2,l3,x0,x1,x2,x3,x[4];
   
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   itest=0;
   l0=NPROC0*L0;
   l1=NPROC1*L1;
   l2=NPROC2*L2;
   l3=NPROC3*L3;
   is=(s[0]+s[1]+s[2]+s[3])&0x1;

   if (my_rank==0)
   {
      for (x0=0;x0<l0;x0++)
      {
         for (x1=0;x1<l1;x1++)
         {
            for (x2=0;x2<l2;x2++)
            {
               for (x3=0;x3<l3;x3++)
               {
                  if (((x0+x1+x2+x3)&0x1)==1)
                  {   
                     x[0]=x0;
                     x[1]=x1;
                     x[2]=x2;
                     x[3]=x3;               
               
                     ipt_global(x,&ip,&ix);
                     ix=(ip-1)*4*VOLUME+8*ix;

                     for (mu=0;mu<4;mu++)
                        x[mu]+=s[mu];

                     if (is==1)
                     {
                        for (mu=0;mu<4;mu++)
                        {
                           x[mu]+=1;
                           ipt_global(x,&iq,&iy);
                           iy=(iq-1)*4*VOLUME+8*iy;               
                           itest+=cmp_su3(utot0+ix+2*mu,utot1+iy+2*mu+1);
                           
                           x[mu]-=2;
                           ipt_global(x,&iq,&iy);
                           iy=(iq-1)*4*VOLUME+8*iy;               
                           itest+=cmp_su3(utot0+ix+2*mu+1,utot1+iy+2*mu);

                           x[mu]+=1;
                        }
                     }
                     else
                     {
                        ipt_global(x,&iq,&iy);
                        iy=(iq-1)*4*VOLUME+8*iy;               
                        
                        for (mu=0;mu<8;mu++)
                           itest+=cmp_su3(utot0+ix+mu,utot1+iy+mu);
                     }
                  }
               }
            }
         }
      }
   }

   return itest;
}


static void random_vec(int *svec)
{
   int mu,bs[4];
   double r[4];

   bs[0]=NPROC0*L0;
   bs[1]=NPROC1*L1;
   bs[2]=NPROC2*L2;
   bs[3]=NPROC3*L3;

   ranlxd(r,4);

   for (mu=0;mu<4;mu++)
   {
      svec[mu]=(int)((double)(bs[mu])*r[mu]);
      if (svec[mu]>(bs[mu]/2))
         svec[mu]-=bs[mu];
   }
   
   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);   
}

               
int main(int argc,char *argv[])
{
   int my_rank,itest;
   int ifc,mu,s[4],n;
   FILE *log=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      log=freopen("check8.log","w",stdout);
      
      printf("\n");
      printf("Translation of the single-precision gauge field\n");
      printf("-----------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }
   
   geometry();
   alloc_u();
   ub=pu[VOLUME/2][0];

   if (my_rank==0)
   {
      utot0=amalloc(8*VOLUME*NPROC*sizeof(su3),ALIGN);
      utot1=utot0+4*VOLUME*NPROC;
      error_root(utot0==NULL,1,"main [check8.c]",
                 "Unable to allocate auxiliary arrays");
   }

   if (my_rank==0)
      printf("Elementary shift vectors:\n\n");

   for (ifc=0;ifc<8;ifc++)
   {
      random_u();
      get_field();
      cpy_field();
               
      s[0]=0;
      s[1]=0;
      s[2]=0;
      s[3]=0;
      mu=ifc/2;

      if ((ifc&0x1)==0)
         s[mu]=1;
      else
         s[mu]=-1;

      shift_u(s);
      get_field();
      itest=cmp_field(s);

      if (my_rank==0)
      {
         printf("Shift vector (% 3d,% 3d,% 3d,% 3d): ",
                s[0],s[1],s[2],s[3]);

         if (itest==0)
            printf("ok\n");
         else
            printf("failed\n");
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Random shift vectors:\n\n");
   }
   
   for (n=0;n<8;n++)
   {
      random_u();
      get_field();
      cpy_field();

      random_vec(s);

      shift_u(s);
      get_field();
      itest=cmp_field(s);

      if (my_rank==0)
      {
         printf("Shift vector (% 3d,% 3d,% 3d,% 3d): ",
                s[0],s[1],s[2],s[3]);

         if (itest==0)
            printf("ok\n");
         else
            printf("failed\n");
      }
   }
  
   error_chk();
   
   if (my_rank==0)
   {
      printf("\n");
      fclose(log);
   }
   
   MPI_Finalize(); 
   exit(0);
}
