
/*******************************************************************************
*
* File run3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Schwarz-preconditioned HMC simulation of two-flavour QCD
*
* Syntax: run3 -i <filename> [-noloc] [-c <filename> [-a]]
*
* For usage instructions see the file README.run3
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "update.h"
#include "block.h"
#include "hmc.h"
#include "global.h"

#define NAME_SIZE 128

typedef struct
{
   int nt,ns,iac,stat_rop;
   int stat_frf,stat_gcr_frf;
   int stat_acf,stat_gcr_acf;
   double dmin,dmax,dact,apl;
} dat_t;


static double link_mls(su3_dble *u,su3_dble *v)
{
   double mls;
   su3_vector_dble *ui,*vi;

   ui=(su3_vector_dble*)(u);
   vi=(su3_vector_dble*)(v);

   mls=
      _vector_prod_re(*(ui  ),*(vi  ))+
      _vector_prod_re(*(ui+1),*(vi+1))+
      _vector_prod_re(*(ui+2),*(vi+2));

   mls=fabs(2.0-(2.0/3.0)*mls);

   return sqrt(mls);
}


static void av_link_mls(double *dmin,double *dmax,double *davg,double *dsig)
{
   int nb,isw,*lk,*lkm;
   int ix,iy;
   double d,dmn,dmx,dav,dsg;
   su3_dble *u,*ub;
   block_t *b,*bm;
   hmc_t *hflds;

   dmn=2.0;
   dmx=0.0;
   dav=0.0;
   dsg=0.0;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;
   hflds=hmcflds();

   for (;b<bm;b++)
   {
      lk=(*hflds).lks;
      lkm=lk+(*hflds).nlks;

      for (;lk<lkm;lk++)
      {
         ix=(*lk)/8;
         iy=(*b).imb[ix+(*b).volh];
         u=pud[iy][0]+(*lk)-8*ix;
         ub=(*b).ud[0]+(*lk);

         d=link_mls(u,ub);

         if (d<dmn)
            dmn=d;
         if (d>dmx)
            dmx=d;

         dav+=d;
         dsg+=d*d;
      }

      hflds+=1;
   }

   MPI_Reduce(&dmn,dmin,1,MPI_DOUBLE,MPI_MIN,0,MPI_COMM_WORLD);
   MPI_Reduce(&dmx,dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&dav,davg,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Reduce(&dsg,dsig,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(dmin,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(dmax,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(davg,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(dsig,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   hflds=hmcflds();
   d=1.0/(double)(NPROC*nb*(*hflds).nlks);
   (*davg)*=d;
   (*dsig)*=d;
   (*dsig)=sqrt((*dsig)-(*davg)*(*davg));
}


int main(int argc,char *argv[])
{
   int my_rank,ii,il,ic,ia,ip,n,nl,ns;
   int idat,icnfg,iend;
   int id,level,seed;
   int bs_hmc[4],bs_sap[4];
   int nth,ntr,wrt_dat,ndat_buf,wrt_cnfg,bc,N0,N1,N2;
   int log_nl,log_icnfg,log_isv;
   int nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf;
   int nkv_sap,nmr_sap,ncy_sap,nvec;
   int stat_acf,stat_gcr_acf;
   int stat_rop,stat_frf,stat_gcr_frf;
   int ifail,sail,iac,siac;
   double beta,kappa,csw,tau;
   double res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf;
   double dmin,dmax,davg,dsig,act0,act1,dact,apl;
   lat_parms_t lat;
   gcr_parms_t gcr;
   hmc_parms_t hmc;
   dat_t *adat=NULL,*ndat;
   char line[NAME_SIZE],log_pgrid[NAME_SIZE],pgrid[NAME_SIZE];
   char log_dir[NAME_SIZE],dat_dir[NAME_SIZE];
   char loc_dir[NAME_SIZE],cnfg_dir[NAME_SIZE];
   char log_file[NAME_SIZE],dat_file[NAME_SIZE];
   char log_save[NAME_SIZE],dat_save[NAME_SIZE];
   char loc_file[NAME_SIZE],cnfg_file[NAME_SIZE],end_file[NAME_SIZE];
   char base[NAME_SIZE],nbase[NAME_SIZE],cnfg[NAME_SIZE],ncnfg[NAME_SIZE];
   FILE *in=NULL,*log=NULL,*dat=NULL,*end=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      log=freopen("MPI_Errors","w",stdout);

      ii=find_opt(argc,argv,"-i");
      error_root((ii==0)||(ii==(argc-1)),1,"main [run3.c]",
                 "Syntax: run3 -i <filename> [-noloc] [-c <filename> [-a]]");

      in=freopen(argv[ii+1],"r",stdin);
      error_root(in==NULL,1,"main [run3.c]","Unable to open input file");

      scanf("id %d\n",&id);      
      sprintf(line,
              "log_dir %%%ds dat_dir %%%ds loc_dir %%%ds cnfg_dir %%%ds",
              NAME_SIZE-1,NAME_SIZE-1,NAME_SIZE-1,NAME_SIZE-1);
      scanf(line,log_dir,dat_dir,loc_dir,cnfg_dir);
      scanf("\n");
      scanf("beta %lf kappa %lf csw %lf tau %lf\n",&beta,&kappa,&csw,&tau);
      scanf("bc %d level %d seed %d\n",&bc,&level,&seed);
      scanf("nth %d ntr %d\n",&nth,&ntr);
      scanf("wrt_dat %d ndat_buf %d wrt_cnfg %d\n",
            &wrt_dat,&ndat_buf,&wrt_cnfg);
      scanf("bs_hmc %d %d %d %d\n",&bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      scanf("N0 %d N1 %d N2 %d\n",&N0,&N1,&N2);
      scanf("nmx_acf %d res_acf %lf\n",&nmx_acf,&res_acf);
      scanf("nmx_gcr_acf %d res_gcr_acf %lf\n",&nmx_gcr_acf,&res_gcr_acf);
      scanf("nmx_rop %d res_rop %lf\n",&nmx_rop,&res_rop);
      scanf("nmx_frf %d res_frf %lf\n",&nmx_frf,&res_frf);
      scanf("nmx_gcr_frf %d res_gcr_frf %lf\n",&nmx_gcr_frf,&res_gcr_frf);
      scanf("bs_sap %d %d %d %d\n",&bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      scanf("nkv_sap %d nmr_sap %d ncy_sap %d",&nkv_sap,&nmr_sap,&ncy_sap);
      fclose(in);

      il=find_opt(argc,argv,"-noloc");      
      ic=find_opt(argc,argv,"-c");
      ia=find_opt(argc,argv,"-a");
      error_root(((ic>0)&&(ic==(argc-1)))||((ia>0)&&(ic==0)),1,"main [run3.c]",
                 "Syntax: run3 -i <filename> [-noloc] [-c <filename> [-a]]");
      if (ic>0)
         strncpy(cnfg,argv[ic+1],NAME_SIZE);
   }

   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(log_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(dat_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&tau,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&level,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&seed,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nth,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ntr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&wrt_dat,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ndat_buf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&wrt_cnfg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&N0,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&N1,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&N2,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&il,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ic,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ia,1,MPI_INT,0,MPI_COMM_WORLD);

   nvec=start_bshift(bs_hmc);   
   check_dir_root(log_dir);
   check_dir_root(dat_dir);
   check_dir_root(cnfg_dir);
   
   if (il==0)
      check_dir(loc_dir);
   
   sprintf(base,"%dx%dx%dx%dp%db%1.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,NPROC,beta,kappa,csw,id);

   sprintf(nbase,"%dx%dx%dx%db%1.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   sprintf(pgrid,"%dx%dx%dx%d",NPROC0,NPROC1,NPROC2,NPROC3);

   if (ic>0)
   {
      MPI_Bcast(cnfg,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
      strcpy(ncnfg,cnfg);

      if (strchr(ncnfg,'p')!=NULL)
         strcpy(strchr(ncnfg,'p'),strchr(cnfg,'b'));

      *(strchr(ncnfg,'n'))='\0';
   }
   
   if (my_rank==0)
   {
      error_root(((strlen(log_dir)+strlen(base)+6)>=NAME_SIZE)||
                 ((strlen(dat_dir)+strlen(base)+6)>=NAME_SIZE),
                 1,"main [run3.c]","log_dir or dat_dir name is too long");
      
      sprintf(log_file,"%s/%s.log",log_dir,base);
      sprintf(dat_file,"%s/%s.dat",dat_dir,base);
      sprintf(end_file,"%s/%s.end",log_dir,base);
      sprintf(log_save,"%s~",log_file);
      sprintf(dat_save,"%s~",dat_file);

      error_root((nth<0)||(nth>ntr)||((nth%(ndat_buf*wrt_dat))!=0)||
                 ((wrt_cnfg%(ndat_buf*wrt_dat))!=0)||(((ntr-nth)%wrt_cnfg)!=0),
                 1,"main [run3.c]",
                 "Improper ntr,nth,wrt_dat,ndat_buf or wrt_cnfg");

      error_root(((nth%nvec)!=0)||((wrt_cnfg%nvec)!=0),1,"main [run3.c]",
                 "nth and wrt_cnfg must be multiples of %s %d",
                 "the block shift cycle",nvec);

      error_root(((wrt_dat%nvec)!=0)&&((nvec%wrt_dat)!=0),1,"main [run3.c]",
                 "wrt_dat must be a multiple or an %s %d",
                 "integral fraction of the block shift cycle",nvec);

      adat=amalloc(ndat_buf*sizeof(dat_t),3);
      error_root(adat==NULL,1,"main [run3.c]","Unable to allocate data array");
      
      if (ia==0)
      {
         end=fopen(log_file,"r");
         dat=fopen(dat_file,"rb");

         error_root((end!=NULL)||(dat!=NULL),1,"main [run3.c]",
                    "Attempt to overwrite old .log or .dat file");

         error_root((ic>0)&&(strcmp(ncnfg,nbase)==0),1,"main [run3.c]",
                    "Run id was already used in a previous run\n%s",
                    "with the same lattice parameters");

         nl=0;
         icnfg=1;
      }
      else
      {
         end=fopen(log_file,"r");
         error_root(end==NULL,1,"main [run3.c]","Unable to open log file");

         *log_pgrid='\0';
         log_nl=0;
         log_isv=0;
         log_icnfg=0;
         
         for (;feof(end)==0;)
         {
            if (fgets(line,NAME_SIZE,end)!=NULL)
            {
               if (strstr(line,"process grid")!=NULL)
                  sscanf(line,"%*s %*s %s",log_pgrid);
               else if (strstr(line,"Trajectory no")!=NULL)
               {
                  sscanf(line,"Trajectory no %d",&log_nl);
                  log_isv=0;
               }
               else if (strstr(line,"Configuration no")!=NULL)
               {
                  sscanf(line,"Configuration no %d",&log_icnfg);
                  log_isv=1;
               }
            }
         }

         fclose(end);         

         dat=fopen(dat_file,"rb");
         error_root(dat==NULL,1,"main [run3.c]","Unable to open data file");

         for (n=0;feof(dat)==0;n++)
            fread(adat,sizeof(dat_t),1,dat);

         fclose(dat);
         
         if (n>=1)
            nl=(*adat).nt;
         else
            nl=0;

         sscanf(strchr(cnfg,'n'),"n%d",&icnfg);
         
         error_root(strstr(ncnfg,nbase)==NULL,1,"main [run3.c]",
                    "Continuation run:\n%s",
                    "Lattice parameters do not match the previous ones");

         error_root(strcmp(log_pgrid,pgrid)!=0,1,"main [run3.c]",
                    "Continuation run:\n%s",
                    "Process grid does not match the previous one");

         error_root(log_icnfg!=icnfg,1,"main [run3.c]",
                    "Continuation run:\n%s %s",
                    "Initial configuration is not the last one of",
                    "the previous run");

         error_root((log_isv==0)||(log_nl!=nl),1,"main [run3.c]",
                    "Continuation run:\n%s",
                    "The log and/or the dat file extends beyond the last save");
         
         icnfg+=1;
      }

      if (il==0)
      {
         sprintf(loc_file,"/n%d_%d",icnfg+(ntr-nth)/wrt_cnfg,NPROC-1);
         error_root((strlen(loc_dir)+strlen(base)+strlen(loc_file))>=NAME_SIZE,
                    1,"main [run3.c]","loc_dir name is too long");
      }
         
      sprintf(cnfg_file,"/n%d",icnfg+(ntr-nth)/wrt_cnfg);
      error_root((strlen(cnfg_dir)+strlen(nbase)+strlen(cnfg_file))>=NAME_SIZE,
                 1,"main [run3.c]","cnfg_dir name is too long");
   }

   MPI_Bcast(&icnfg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nl,1,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(level,seed^nl);
   geometry();
   alloc_ud();
   alloc_s(2*nkv_sap+1);
   alloc_sd(4);

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (ic>0)
   {
      if (strchr(cnfg,'p')==NULL)
      {
         sprintf(cnfg_file,"%s/%s",cnfg_dir,cnfg);
         import_cnfg(cnfg_file);
      }
      else
      {
         error(il>0,1,"main [run3.c]",
               "Attempt to read imported configuration when -noloc is set");
         
         sprintf(loc_file,"%s/%s_%d",loc_dir,cnfg,my_rank);
         read_cnfg(loc_file);
      }
   }
   else
   {
      random_ud();
      error_chk();
   }

   if (my_rank==0)
   {
      ip=ftell(log);
      fclose(log);

      if (ip==0)
         remove("MPI_Errors");

      if (ia==0)
         log=freopen(log_file,"w",stdout);
      else
         log=freopen(log_file,"a",stdout);

      error_root(log==NULL,1,"main [run3.c]","Unable to open log file");

      if (ia==0)
      {
         printf("\n");
         printf("Generation of gauge field configurations\n");
         printf("----------------------------------------\n\n");

         if (ic>0)
            printf("New run, start from configuration %s\n\n",cnfg);
         else
            printf("New run, start from random configuration\n\n");
      }
      else
         printf("Continuation run, start from configuration %s\n\n",cnfg);

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      printf("O(a) improved theory\n");
      printf("beta = %.6f\n",lat.beta);
      printf("kappa = %.6f\n",lat.kappa);
      printf("csw = %.6f\n",lat.csw);
      printf("tau = %.6f\n",hmc.tau);
      printf("bc = %+d\n",hmc.bc);

      if ((ia>0)&&(strchr(cnfg,'p')!=NULL))
      {
         printf("State of ranlxs and ranlxd read from\n");
         printf("initial field-configuration file\n\n");
      }
      else
         printf("level = %d, seed = %d, effective seed = %d\n\n",
                level,seed,seed^nl);

      printf("nth = %d, ntr = %d\n",nth,ntr);
      printf("wrt_dat = %d, ndat_buf = %d, wrt_cnfg = %d\n\n",
             wrt_dat,ndat_buf,wrt_cnfg);

      printf("bs_hmc = %d %d %d %d, ",
             bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("block shift cycle = %d\n",nvec);
      printf("N0,N1,N2 = %d,%d,%d\n",hmc.N0,hmc.N1,hmc.N2);
      printf("nmx_rop = %d, res_rop = %.1e\n",hmc.nmx_rop,hmc.res_rop);
      printf("nmx_acf = %d, res_acf = %.1e\n",hmc.nmx_acf,hmc.res_acf);
      printf("nmx_gcr_acf = %d, res_gcr_acf = %.1e\n",
             hmc.nmx_gcr_acf,hmc.res_gcr_acf);
      printf("nmx_frf = %d, res_frf = %.1e\n",hmc.nmx_frf,hmc.res_frf);
      printf("nmx_gcr_frf = %d, res_gcr_frf = %.1e\n",
             hmc.nmx_gcr_frf,hmc.res_gcr_frf);
      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nkv_sap,nmr_sap,ncy_sap = %d,%d,%d\n\n",
             gcr.nkv,gcr.nmr,gcr.ncy);
      if (il>0)
         printf("The local disks are not used\n\n");

      fflush(log);
   }

   idat=0;
   iend=0;
   sail=0;
   siac=0;

   for (n=0;(iend==0)&&(n<ntr);n++)
   {
      ns=mvfld();
      act0=start_hmd(&stat_rop,&ifail);

      if ((stat_rop<0)||(ifail!=0))
      {
         sail+=1;
         n-=1;
         continue;
      }

      act1=run_hmd(&stat_acf,&stat_gcr_acf,&stat_frf,&stat_gcr_frf,&ifail);

      if ((stat_acf<0)||(stat_gcr_acf<0)||
          (stat_frf<0)||(stat_gcr_frf<0)||(ifail!=0))
      {
         sail+=1;
         n-=1;
         continue;
      }      

      iac=accept_step(act0,act1);
      siac+=iac;
      
      if (((ntr-n-1)%wrt_dat)==0)
      {
         dact=act1-act0;
         av_link_mls(&dmin,&dmax,&davg,&dsig);
         apl=plaq_sum_dble()/(double)(6*VOLUME*NPROC);
         
         if (my_rank==0)
         {
            printf("Trajectory no %d\n",nl+n+1);
            printf("start_hmd: ");
            printf("nshift = %d, ",ns);
            printf("stat_rop = %d\n",stat_rop);

            printf("run_hmd: ");
            printf("stat_frf = %d, ",stat_frf);
            printf("stat_gcr_frf = %d, ",stat_gcr_frf);
            printf("stat_acf = %d, ",stat_acf);
            printf("stat_gcr_acf = %d\n",stat_gcr_acf);

            printf("accept_step: ");
            printf("dH = %+.1e, ",dact);
            printf("iac = %d\n",iac);
            printf("Link mileage = %.1e ... %.1e, ",dmin,dmax);
            printf("average = %.1e (%.1e)\n",davg,dsig);
            printf("Average plaquette = %.6f\n\n",apl);

            ndat=adat+idat;
            (*ndat).nt=nl+n+1;
            (*ndat).ns=ns;
            (*ndat).iac=iac;
            (*ndat).stat_rop=stat_rop;
            (*ndat).stat_frf=stat_frf;
            (*ndat).stat_gcr_frf=stat_gcr_frf;
            (*ndat).stat_acf=stat_acf;
            (*ndat).stat_gcr_acf=stat_gcr_acf;
            (*ndat).dmin=dmin;
            (*ndat).dmax=dmax;
            (*ndat).dact=dact;
            (*ndat).apl=apl;
            idat+=1;

            if (idat==ndat_buf)
            {
               dat=fopen(dat_file,"ab");
               error_root(dat==NULL,1,"main [run3.c]",
                          "Unable to open data file");
               fwrite(adat,sizeof(dat_t),ndat_buf,dat);
               error_root(ferror(dat)!=0,1,"main [run3.c]",
                          "Error while saving data to dat file");
               fclose(dat);

               printf("%d data items saved\n",ndat_buf);
               printf("Acceptance rate = %1.2f, ",
                      (double)(siac)/(double)(n+1));
               printf("failed trajectories = %d\n",sail);

               if (((n+1)<nth)||(((ntr-n-1)%wrt_cnfg)!=0))
                  printf("\n");
               
               idat=0;
            }

            fflush(log);
         }
      }

      if (((n+1)>=nth)&&(((ntr-n-1)%wrt_cnfg)==0))
      {
         if (il==0)
         {
            sprintf(loc_file,"%s/%sn%d_%d",loc_dir,base,icnfg,my_rank);
            write_cnfg(loc_file);
         }
         
         sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
         export_cnfg(cnfg_file);

         if (my_rank==0)
         {
            if (il==0)
            {
               printf("Configuration no %d saved on the local disks ",icnfg);
               printf("and exported\n\n");
            }
            else
               printf("Configuration no %d exported\n\n",icnfg);

            end=fopen(end_file,"r");

            if (end!=NULL)
            {
               fclose(end);
               remove(end_file);
               iend=1;
               printf("End flag set, run stopped\n\n");
            }

            fflush(log);
         }

         error_chk();
         icnfg+=1;
         MPI_Bcast(&iend,1,MPI_INT,0,MPI_COMM_WORLD);

         if (my_rank==0)
         {
            copy_file(log_file,log_save);
            copy_file(dat_file,dat_save);
         }
      }
   }

   if (my_rank==0)
      fclose(log);

   MPI_Finalize();
   exit(0);
}
