
/*******************************************************************************
*
* File parms.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Parameter data base
*
* The externally accessible functions are
*
*   lat_parms_t set_lat_parms(double beta,double kappa,double csw)
*     Sets the basic lattice parameters. The return value is a structure
*     that contains the lattice parameters
*
*   lat_parms_t lat_parms(void)
*     Returns the current lattice parameters in a structure with elements
*     *.beta,*.kappa,*.m0,*.csw where m0 is bare quark mass determined by
*     kappa
*
*   gcr_parms_t set_gcr_parms(int *bs,int nmr,int ncy,int nkv)
*     Sets the parameters for the Schwarz-preconditioned GCR solver and
*     allocates (or reallocates) the GCR_BLOCKS grid. Here bs[4] is the
*     block size, nmr the number of block MR iterations, ncy the number
*     of Schwarz cycles and nkv the maximal number of Krylov vectors to
*     be used before the GCR algorithm is restarted. The return value is
*     a structure that contains the solver parameters
*
*   gcr_parms_t gcr_parms(void)
*     Returns the parameters currently set for the Schwarz-preconditioned
*     GCR solver in a structure with elements *.nb,*.nmr,*.ncy,*.nkv where
*     nb is the number of blocks per process in the current GCR_BLOCKS grid
*
*   hmc_parms_t set_hmc_parms(int *bs,int bc,int N0,int N1,int N2,
*                             int nmx_acf,int nmx_gcr_acf,int nmx_rop,
*                             int nmx_frf,int nmx_gcr_frf,
*                             double tau,
*                             double res_acf,double res_gcr_acf,double res_rop,
*                             double res_frf,double res_gcr_frf)
*     Sets the parameters of the Schwarz-preconditioned HMC algorithm and
*     and allocates (or reallocates) the HMC_BLOCKS grid. The parameters are
*
*       bs[4]         Block size of the blocks in the HMC_BLOCKS grid
*
*       bc            Sets the boundary conditions of the quark fields in the
*                     time direction (bc=1: periodic, bc=-1: antiperiodic) 
*
*       N0,N1,N2      Integration step numbers (as in the CPC paper)
*
*       tau           Trajectory length
*
*       nmx_acf       Maximal solver iteration number and required relative
*       res_acf       residue when calculating the block pseudo-fermion action
*
*       nmx_gcr_acf   Same for the calculation of the block-interaction
*       res_gcr_acf   pseudo-fermion action ("gcr" = global correction)
*
*       nmx_rop       Same for the evaluation of the R-operator, when
*       res_rop       calculating the block-interaction pseudo-fermion field
*                     at the beginning of the trajectory
*
*       nmx_frf       Same for the calculation of block force         
*       res_frf 
*
*       nmx_gcr_frf   Same for the calculation of the block-interaction force
*       res_gcr_frf
*
*     The return value is a structure that contains the HMC parameters
*
*   hmc_parms_t hmc_parms(void)
*     Returns the parameters currently set for the Schwarz-preconditioned
*     HMC algorithm in a structure with elements *.nb,*.bc,*.N0,*.N1,..,
*     where nb is the number of blocks per process in the HMC_BLOCKS grid.
*     The other elements match the arguments of set_hmc_parms() apart from
*     the block size bs which is omitted
*
* Notes:
*
* To ensure the consistency of the data base, the parameters must be set
* simultaneously on all processes. The types lat_parms_t, gcr_parms_t and
* hmc_parms_t are defined in the file flags.h
*
*******************************************************************************/

#define PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "start.h"
#include "flags.h"
#include "block.h"
#include "global.h"

static lat_parms_t lat={0.0,0.125,0.0,0.0};
static gcr_parms_t gcr={0,0,0,0};
static hmc_parms_t hmc={0,1,1,1,1,1,1,1,1,1,0.0,1.0,1.0,1.0,1.0,1.0};


lat_parms_t set_lat_parms(double beta,double kappa,double csw)
{
   double dprms[3];

   if (NPROC>1)
   {
      dprms[0]=beta;
      dprms[1]=kappa;
      dprms[2]=csw;

      MPI_Bcast(dprms,3,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((dprms[0]!=beta)||(dprms[1]!=kappa)||(dprms[2]!=csw),1,
             "set_lat_parms [parms.c]","Parameters are not global");
   }

   error_root(kappa<=0.0,1,
              "set_lat_parms [parms.c]","kappa must be positive");

   if ((lat.beta!=beta)||(lat.kappa!=kappa)||(lat.csw!=csw))
      set_flags(CHANGED_LAT_PARMS);

   lat.beta=beta;
   lat.kappa=kappa;
   lat.m0=1.0/(2.0*kappa)-4.0;
   lat.csw=csw;

   return lat;
}


lat_parms_t lat_parms(void)
{
   return lat;
}


gcr_parms_t set_gcr_parms(int *bs,int nmr,int ncy,int nkv)
{
   int iprms[3],nb,isw;

   if (NPROC>1)
   {
      iprms[0]=nmr;
      iprms[1]=ncy;
      iprms[2]=nkv;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=nmr)||(iprms[1]!=ncy)||(iprms[2]!=nkv),1,
             "set_gcr_parms [parms.c]","Parameters are not global");
   }

   error_root((nmr<0)||(ncy<0)||(nkv<1),1,
               "set_gcr_parms [parms.c]","Improper parameters nmr,ncy,nkv");
   alloc_bgr(GCR_BLOCKS,bs);
   blk_list(GCR_BLOCKS,&nb,&isw);

   gcr.nb=nb;
   gcr.nmr=nmr;
   gcr.ncy=ncy;
   gcr.nkv=nkv;

   return gcr;
}


gcr_parms_t gcr_parms(void)
{
   int nb,isw;

   blk_list(GCR_BLOCKS,&nb,&isw);

   if (nb==0)
   {
      gcr.nmr=0;
      gcr.ncy=0;
      gcr.nkv=0;
   }

   gcr.nb=nb;

   return gcr;
}


hmc_parms_t set_hmc_parms(int *bs,int bc,int N0,int N1,int N2,
                          int nmx_acf,int nmx_gcr_acf,int nmx_rop,
                          int nmx_frf,int nmx_gcr_frf,
                          double tau,
                          double res_acf,double res_gcr_acf,double res_rop,
                          double res_frf,double res_gcr_frf)
{
   int iprms[9],nb,isw;
   double dprms[6];

   if (NPROC>1)
   {
      iprms[0]=bc;
      iprms[1]=N0;
      iprms[2]=N1;
      iprms[3]=N2;      
      iprms[4]=nmx_acf;
      iprms[5]=nmx_gcr_acf;
      iprms[6]=nmx_rop;
      iprms[7]=nmx_frf;      
      iprms[8]=nmx_gcr_frf;

      dprms[0]=tau;
      dprms[1]=res_acf;
      dprms[2]=res_gcr_acf;
      dprms[3]=res_rop;
      dprms[4]=res_frf;
      dprms[5]=res_gcr_frf;
      
      MPI_Bcast(iprms,9,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,6,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=bc)||(iprms[1]!=N0)||(iprms[2]!=N1)||(iprms[3]!=N2)||
            (iprms[4]!=nmx_acf)||(iprms[5]!=nmx_gcr_acf)||(iprms[6]!=nmx_rop)||
            (iprms[7]!=nmx_frf)||(iprms[8]!=nmx_gcr_frf)||(dprms[0]!=tau)||
            (dprms[1]!=res_acf)||(dprms[2]!=res_gcr_acf)||(dprms[3]!=res_rop)||
            (dprms[4]!=res_frf)||(dprms[5]!=res_gcr_frf),1,
            "set_hmc_parms [parms.c]","Parameters are not global");
   }

   error_root((abs(bc)!=1)||(N0<1)||(N1<1)||(N2<1),1,
              "set_hmc_parms [parms.c]","Improper parameters bc,N0,N1,N2");
   error_root((nmx_acf<1)||(nmx_gcr_acf<1)||(nmx_rop<1)||
              (nmx_frf<1)||(nmx_gcr_frf<1),1,
              "set_hmc_parms [parms.c]","Improper parameters nmx_*");
   error_root((res_acf<=DBL_EPSILON)||(res_gcr_acf<=DBL_EPSILON)||
              (res_rop<=DBL_EPSILON)||(res_frf<=DBL_EPSILON)||
              (res_gcr_frf<=DBL_EPSILON),1,
              "set_hmc_parms [parms.c]","Improper parameters res_*");
   
   alloc_bgr(HMC_BLOCKS,bs);
   blk_list(HMC_BLOCKS,&nb,&isw);

   hmc.nb=nb;
   hmc.bc=bc;
   hmc.N0=N0;
   hmc.N1=N1;
   hmc.N2=N2;
   hmc.nmx_acf=nmx_acf;
   hmc.nmx_gcr_acf=nmx_gcr_acf;
   hmc.nmx_rop=nmx_rop;   
   hmc.nmx_frf=nmx_frf;
   hmc.nmx_gcr_frf=nmx_gcr_frf;
   hmc.tau=tau;
   hmc.res_acf=res_acf;
   hmc.res_gcr_acf=res_gcr_acf;   
   hmc.res_rop=res_rop;
   hmc.res_frf=res_frf;
   hmc.res_gcr_frf=res_gcr_frf;

   return hmc;
}


hmc_parms_t hmc_parms(void)
{
   int nb,isw;

   blk_list(HMC_BLOCKS,&nb,&isw);

   if (nb==0)
   {
      hmc.bc=1;
      hmc.N0=0;
      hmc.N1=0;
      hmc.N2=0;
      hmc.nmx_acf=1;
      hmc.nmx_gcr_acf=1;
      hmc.nmx_rop=1;
      hmc.nmx_frf=1;
      hmc.nmx_gcr_frf=1;
      hmc.tau=0.0;
      hmc.res_acf=1.0;
      hmc.res_gcr_acf=1.0;
      hmc.res_rop=1.0;
      hmc.res_frf=1.0;
      hmc.res_gcr_frf=1.0;
   }

   hmc.nb=nb;

   return hmc;
}
