
/*******************************************************************************
*
* File wspace.c
*
* Copyright (C) 2006 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Workspace administration
*
* The externally accessible functions are
*
*   spinor **reserve_ws(int ns)
*     Reserves a new workspace of ns global single-precision spinor fields
*     and returns the array ws[0],..,ws[ns-1] of the base addresses of the
*     fields in the workspace. No workspace is reserved and a NULL pointer
*     is returned if ns<=0
*
*   int release_ws(void)
*     Releases the workspace of global single-precision spinor fields that
*     was last reserved and returns the number of fields that are released
*
*   int ws_size(void)
*     Returns the number of global single-precision spinor fields that
*     are currently reserved
*
*   spinor_dble **reserve_wsd(int ns)
*     Reserves a new workspace of ns global double-precision spinor fields
*     and returns the array wsd[0],..,wsd[ns-1] of the base addresses of the
*     fields in the workspace. No workspace is reserved and a NULL pointer
*     is returned if ns<=0
*
*   int release_wsd(void)
*     Releases the workspace of global double-precision spinor fields that
*     was last reserved and returns the number of fields that are released
*
*   int wsd_size(void)
*     Returns the number of global double-precision spinor fields that
*     are currently reserved
*
* Notes:
*
* By definition a workspace is a set of global single- or double-precision
* spinor fields which is reserved for the current program. This module allows
* to assign and release the required workspaces dynamically in such a way
* that the privacy of the workspaces is guaranteed
*
* The total workspace is the set of all allocated spinor fields. All programs
* (even the main program) must reserve the required workspace and they must
* release the workspace before the program returns to the calling program
*
* It is possible to reserve several workspaces in a given program. In this
* case, a corresponding number of calls to ws_release and wsd_release is
* required to release all the workspaces that were reserved
*
*******************************************************************************/

#define WSPACE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "flags.h"
#include "global.h"

static int nst=0,iws=0,nwst=0,*nws;
static int nsdt=0,iwsd=0,nwsdt=0,*nwsd;
static spinor **ws;
static spinor_dble **wsd;


static void alloc_nws(void)
{
   int n;

   error_root(nwst!=0,1,"alloc_nws [wspace.c]","Fields are in use");
   
   if (nst>0)
   {
      afree(nws);
      afree(ws);
      nws=NULL;
      ws=NULL;
   }

   nst=no_s;
   iws=0;
   nwst=0;

   if (nst>0)
   {
      nws=amalloc(nst*sizeof(int),3);
      ws=amalloc(nst*sizeof(*ws),3);

      error((nws==NULL)||(ws==NULL),1,"alloc_nws [wspace.c]",
             "Unable to allocated index arrays");

      for (n=0;n<nst;n++)
      {
         nws[n]=0;
         ws[n]=NULL;
      }
   }
}


static void alloc_nwsd(void)
{
   int n;

   error_root(nwsdt!=0,1,"alloc_nwsd [wspace.c]","Fields are in use");   

   if (nsdt>0)
   {
      afree(nwsd);
      afree(wsd);
      nwsd=NULL;
      wsd=NULL;
   }

   nsdt=no_sd;
   iwsd=0;
   nwsdt=0;

   if (nsdt>0)
   {
      nwsd=amalloc(nsdt*sizeof(int),3);
      wsd=amalloc(nsdt*sizeof(*wsd),3);

      error((nwsd==NULL)||(wsd==NULL),1,"alloc_nwsd [wspace.c]",
             "Unable to allocated index arrays");

      for (n=0;n<nsdt;n++)
      {
         nwsd[n]=0;
         wsd[n]=NULL;
      }
   }
}


spinor **reserve_ws(int ns)
{
   int iprms[1],n,i;

   if (NPROC>1)
   {
      iprms[0]=ns;

      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);

      error(iprms[0]!=ns,1,"reserve_ws [wspace.c]",
            "Parameters are not global");
   }

   if (nst!=no_s)
      alloc_nws();

   if (ns>0)
   {
      error((nwst+ns)>nst,1,"reserve_ws [wspace.c]",
            "Requested too many fields (tot=%d,use=%d,req=%d)",nst,nwst,ns);

      n=nwst;
      nws[iws]=ns;
      nwst+=ns;
      iws+=1;

      for (i=n;i<(n+ns);i++)
         ws[i]=ps[i][0];

      return ws+n;
   }
   else
      return NULL;
}


int release_ws(void)
{
   int ns,i;

   if (nwst==0)
      return 0;
   else
   {
      iws-=1;
      ns=nws[iws];
      nwst-=ns;
      nws[iws]=0;

      for (i=nwst;i<(nwst+ns);i++)
         ws[i]=NULL;
      
      return ns;
   }
}


int ws_size(void)
{
   return nwst;
}


spinor_dble **reserve_wsd(int ns)
{
   int iprms[1],n,i;

   if (NPROC>1)
   {
      iprms[0]=ns;

      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);

      error(iprms[0]!=ns,1,"reserve_wsd [wspace.c]",
            "Parameters are not global");
   }

   if (nsdt!=no_sd)
      alloc_nwsd();

   if (ns>0)
   {
      error((nwsdt+ns)>nsdt,1,"reserve_wsd [wspace.c]",
            "Requested too many fields (tot=%d,use=%d,req=%d)",nsdt,nwsdt,ns);

      n=nwsdt;
      nwsd[iwsd]=ns;
      nwsdt+=ns;
      iwsd+=1;

      for (i=n;i<(n+ns);i++)
         wsd[i]=psd[i][0];
      
      return wsd+n;
   }
   else
      return NULL;
}


int release_wsd(void)
{
   int ns,i;

   if (nwsdt==0)
      return 0;
   else
   {
      iwsd-=1;
      ns=nwsd[iwsd];
      nwsdt-=ns;
      nwsd[iwsd]=0;

      for (i=nwsdt;i<(nwsdt+ns);i++)
         wsd[i]=NULL;
      
      return ns;
   }
}


int wsd_size(void)
{
   return nwsdt;
}

