
/*******************************************************************************
*
* File bndop.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs related to the operator R_2 and the associated pseudo-fermion
* field chi=phi_2 on the black blocks of the HMC_BLOCKS grid 
*
* The externally accessible functions are
*
*   void random_chi(void)
*     Sets the pseudo-fermion field chi to random values with distribution
*     proportional to exp{-(chi,chi)}. The field is generated on all black
*     blocks and satisfies P_dOmega^star*chi=chi
*
*   void assign_chi2sd(int k)
*     Assigns the pseudo-fermion field chi to the global spinor field
*     psd[k], setting all spinors that are not on the interior boundary
*     of the black blocks to zero
*
*   void assign_sd2chi(int k)
*     Assigns P_dOmega^star*psd[k] to the pseudo-fermion field chi
*
*   void sub_assign_PdOs(int k,int l)
*     Replaces the global spinor field psd[k] by psd[k]-P_dOmega^star*psd[l]
*
*   void ROp(int *status,int *ifail)
*     Replaces the pseudo-fermion field chi by R_2*chi, where R_2 is the
*     block interaction operator, the gauge field being the current double-
*     precision field on the blocks. The swb arrays must be up-to-date and
*     the SW-term on the blocks is recalculated and inverted when needed.
*      On exit status reports the average number of BiCGstab iterations that
*     were required to invert the block operators, or a negative number if
*     the solver failed. The parameter ifail is set to 0 or 1 depending on
*     whether the inversions of the SW-term (if any) were safe or not
*
* Notes:
*
* The definitions of the projector P_dOmega^star and the operator R_2 are
* as in CPC 165 (2005) 199. All programs in this file act globally and must
* be called simultaneously on all processes
*
* The program ROp() needs a workspace of 2 global double-precision spinor
* fields that is reserved using the utilities in the module flags/wspace.c.
* The solver and any other parameters are obtained from the parameter data
* base (module flags/parms.c)
*
*******************************************************************************/

#define BNDOP_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "random.h"
#include "linalg.h"
#include "linsolv.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "global.h"

#define _vector_m_mul(r,s) \
   (r).c1.re=-(s).c1.re; \
   (r).c1.im=-(s).c1.im; \
   (r).c2.re=-(s).c2.re; \
   (r).c2.im=-(s).c2.im; \
   (r).c3.re=-(s).c3.re; \
   (r).c3.im=-(s).c3.im

#define _vector_i_mul(r,s) \
   (r).c1.re=-(s).c1.im; \
   (r).c1.im= (s).c1.re; \
   (r).c2.re=-(s).c2.im; \
   (r).c2.im= (s).c2.re; \
   (r).c3.re=-(s).c3.im; \
   (r).c3.im= (s).c3.re

#define _vector_mi_mul(r,s) \
   (r).c1.re= (s).c1.im; \
   (r).c1.im=-(s).c1.re; \
   (r).c2.re= (s).c2.im; \
   (r).c2.im=-(s).c2.re; \
   (r).c3.re= (s).c3.im; \
   (r).c3.im=-(s).c3.re

#define _vector_mul_assign(r,c) \
   (r).c1.re*=(c); \
   (r).c1.im*=(c); \
   (r).c2.re*=(c); \
   (r).c2.im*=(c); \
   (r).c3.re*=(c); \
   (r).c3.im*=(c)

static spinor_dble rs;
static block_t *bsv;


static void pr_sd2sd(int ifc,spinor_dble *s,spinor_dble *r)
{
   switch (ifc)
   {
      case 0:
         _vector_add((*r).c1,(*s).c1,(*s).c3);
         _vector_mul_assign((*r).c1,0.5);
         (*r).c3=(*r).c1;
         _vector_add((*r).c2,(*s).c2,(*s).c4);
         _vector_mul_assign((*r).c2,0.5);
         (*r).c4=(*r).c2;
         break;
      case 1:
         _vector_sub((*r).c1,(*s).c1,(*s).c3);
         _vector_mul_assign((*r).c1,0.5);
         _vector_m_mul((*r).c3,(*r).c1);
         _vector_sub((*r).c2,(*s).c2,(*s).c4);
         _vector_mul_assign((*r).c2,0.5);
         _vector_m_mul((*r).c4,(*r).c2);
         break;
      case 2:
         _vector_i_add((*r).c1,(*s).c1,(*s).c4);
         _vector_mul_assign((*r).c1,0.5);
         _vector_mi_mul((*r).c4,(*r).c1);
         _vector_i_add((*r).c2,(*s).c2,(*s).c3);         
         _vector_mul_assign((*r).c2,0.5);
         _vector_mi_mul((*r).c3,(*r).c2);
         break;
      case 3:
         _vector_i_sub((*r).c1,(*s).c1,(*s).c4);
         _vector_mul_assign((*r).c1,0.5);
         _vector_i_mul((*r).c4,(*r).c1);
         _vector_i_sub((*r).c2,(*s).c2,(*s).c3);
         _vector_mul_assign((*r).c2,0.5);
         _vector_i_mul((*r).c3,(*r).c2);
         break;
      case 4:
         _vector_add((*r).c1,(*s).c1,(*s).c4);
         _vector_mul_assign((*r).c1,0.5);
         (*r).c4=(*r).c1;         
         _vector_sub((*r).c2,(*s).c2,(*s).c3);
         _vector_mul_assign((*r).c2,0.5);
         _vector_m_mul((*r).c3,(*r).c2);
         break;
      case 5:
         _vector_sub((*r).c1,(*s).c1,(*s).c4);
         _vector_mul_assign((*r).c1,0.5);
         _vector_m_mul((*r).c4,(*r).c1);
         _vector_add((*r).c2,(*s).c2,(*s).c3);
         _vector_mul_assign((*r).c2,0.5);
         (*r).c3=(*r).c2;   
         break;
      case 6:
         _vector_i_add((*r).c1,(*s).c1,(*s).c3);
         _vector_mul_assign((*r).c1,0.5);
         _vector_mi_mul((*r).c3,(*r).c1);
         _vector_i_sub((*r).c2,(*s).c2,(*s).c4);
         _vector_mul_assign((*r).c2,0.5);
         _vector_i_mul((*r).c4,(*r).c2);
         break;
      case 7:
         _vector_i_sub((*r).c1,(*s).c1,(*s).c3);
         _vector_mul_assign((*r).c1,0.5);
         _vector_i_mul((*r).c3,(*r).c1);
         _vector_i_add((*r).c2,(*s).c2,(*s).c4);
         _vector_mul_assign((*r).c2,0.5);
         _vector_mi_mul((*r).c4,(*r).c2);
         break;
      default:
         *r=*s;
         break;
   }
}
   

void random_chi(void)
{
   int nb,nbh,isw,*ifc;
   spinor_dble *s,*sm;
   hmc_t *hflds,*hfldsm;

   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   error(hflds==NULL,1,"random_chi [bndop.c]",
         "HMC fields are not allocated");

   nbh=nb/2;
      
   if (isw==1)
      hflds+=nbh;

   hfldsm=hflds+nbh;

   for (;hflds<hfldsm;hflds++)
   {
      s=(*hflds).chi;
      random_sd((*hflds).nbnd,s,1.0);

      ifc=(*hflds).ifc;      
      sm=s+(*hflds).nbnd;

      for (;s<sm;s++)
      {
         if ((*ifc)<8)
            pr_sd2sd(*ifc,s,s);
         ifc+=1;
      }
   }
}


void assign_chi2sd(int k)
{
   int nb,nbh,isw,*imb,iprms[1];
   spinor_dble *s,*sk,*sm;
   hmc_t *hflds,*hfldsm;

   iprms[0]=k;

   MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);

   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   error((hflds==NULL)||(k!=iprms[0]),1,"assign_chi2sd [bndop.c]",
         "HMC fields are not allocated or argument is not global");

   error_root((k<0)||(k>=no_sd),1,"assign_chi2sd [bndop.c]",
              "Argument out of range");

   sk=psd[k][0];   
   set_sd2zero(VOLUME,sk);   
   
   nbh=nb/2;
      
   if (isw==1)
      hflds+=nbh;

   hfldsm=hflds+nbh;

   for (;hflds<hfldsm;hflds++)
   {
      imb=(*hflds).imb;
      s=(*hflds).chi;
      sm=s+(*hflds).nbnd;

      for (;s<sm;s++)
      {
         *(sk+(*imb))=*s;
         imb+=1;
      }
   }   
}


void assign_sd2chi(int k)
{
   int nb,nbh,isw,*imb,*ifc,iprms[1];
   spinor_dble *s,*sk,*sm;
   hmc_t *hflds,*hfldsm;

   iprms[0]=k;

   MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);

   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   error((hflds==NULL)||(k!=iprms[0]),1,"assign_sd2chi [bndop.c]",
         "HMC fields are not allocated or argument is not global");

   error_root((k<0)||(k>=no_sd),1,"assign_sd2chi [bndop.c]",
              "Argument out of range");

   sk=psd[k][0];
   nbh=nb/2;
      
   if (isw==1)
      hflds+=nbh;

   hfldsm=hflds+nbh;

   for (;hflds<hfldsm;hflds++)
   {
      imb=(*hflds).imb;
      ifc=(*hflds).ifc;

      s=(*hflds).chi;
      sm=s+(*hflds).nbnd;

      for (;s<sm;s++)
      {
         pr_sd2sd(*ifc,sk+(*imb),s);
         imb+=1;
         ifc+=1;
      }
   }
}


void sub_assign_PdOs(int k,int l)
{
   int nb,nbh,isw,*imb,*ifc,*ifcm,iprms[2];
   spinor_dble *sk,*sl,*rk;
   hmc_t *hflds,*hfldsm;

   iprms[0]=k;
   iprms[1]=l;

   MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   error((hflds==NULL)||(k!=iprms[0])||(l!=iprms[1]),1,
         "sub_assign_PdOs [bndop.c]",
         "HMC fields are not allocated or arguments are not global");

   error_root((k<0)||(k>=no_sd)||(l<0)||(l>=no_sd),1,
              "sub_assign_PdOs [bndop.c]","Arguments out of range");

   sk=psd[k][0];
   sl=psd[l][0];   
   nbh=nb/2;
      
   if (isw==1)
      hflds+=nbh;

   hfldsm=hflds+nbh;

   for (;hflds<hfldsm;hflds++)
   {
      imb=(*hflds).imb;
      ifc=(*hflds).ifc;
      ifcm=ifc+(*hflds).nbnd;

      for (;ifc<ifcm;ifc++)
      {
         pr_sd2sd(*ifc,sl+(*imb),&rs);
         rk=sk+(*imb);
         imb+=1;
         
         _vector_sub_assign((*rk).c1,rs.c1);
         _vector_sub_assign((*rk).c2,rs.c2);
         _vector_sub_assign((*rk).c3,rs.c3);
         _vector_sub_assign((*rk).c4,rs.c4);
      }
   }
}


static void DdOs(int isw,int k,int l)
{
   int nb,nbh,isw0;
   block_t *b,*bm;

   b=blk_list(HMC_BLOCKS,&nb,&isw0);   
   nbh=nb/2;
   bm=b+nbh;
   
   for (;b<bm;b++)
   {
      if (isw==isw0)
         assign_sd2sdb(b,ALL_PTS,k,0);
      else
         set_sd2zero((*b).vol,(*b).sd[0]);

      Qbnds_dble(b,0,0);      
      mulg5_sub_assign_sdbb2sd(b,0,l);
   }

   bm+=nbh;
   
   for (;b<bm;b++)
   {
      if (isw!=isw0)
         assign_sd2sdb(b,ALL_PTS,k,0);
      else
         set_sd2zero((*b).vol,(*b).sd[0]);

      Qbnds_dble(b,0,0);
      mulg5_sub_assign_sdbb2sd(b,0,l);
   }    
}


static void Op_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l;

   k=(pk-(*bsv).sd[0])/((*bsv).vol+1);
   l=(pl-(*bsv).sd[0])/((*bsv).vol+1);
   
   Qhat_blk_dble(bsv,k,l);
   mulg5_dble((*bsv).volh,pl);
}


static void DOinv(int isw,int k,int l,int *status)
{
   int nb,nbh,isw0,stat,vol,volh;
   spinor_dble *se,*so,*re,*ro,*sm;
   block_t *b,*bm;
   hmc_parms_t hmc;

   hmc=hmc_parms();
   b=blk_list(HMC_BLOCKS,&nb,&isw0);   
   nbh=nb/2;

   if (isw!=isw0)
      b+=nbh;

   bm=b+nbh;   

   for (;b<bm;b++)
   {
      bsv=b;
      vol=(*b).vol;
      volh=(*b).volh;
      
      se=(*b).sd[0];
      re=(*b).sd[1];
      so=se+volh;
      ro=re+volh;

      assign_sd2sdb(b,ALL_PTS,k,1);

      mulg5_dble(vol,re);
      apply_sw_dble(volh,(*b).swd+vol,ro,ro);
      Qeo_blk_dble(b,1,1);
      mulg5_dble(volh,re);

      bicg((*b).volh,0,Op_dble,(*b).sd+2,hmc.nmx_rop,hmc.res_rop,re,se,&stat);

      if (((*status)>=0)&&(stat>=0))
         (*status)+=stat;
      else if (stat<0)
         (*status)=stat;

      Qoe_blk_dble(b,0,0);
      apply_sw_dble(volh,(*b).swd+vol,so,so);

      sm=so+volh;

      for (;so<sm;so++)
      {
         _vector_sub((*so).c1,(*ro).c1,(*so).c1);
         _vector_sub((*so).c2,(*ro).c2,(*so).c2);
         _vector_sub((*so).c3,(*ro).c3,(*so).c3);
         _vector_sub((*so).c4,(*ro).c4,(*so).c4);
         ro+=1;
      }
      
      assign_sdb2sd(b,ALL_PTS,0,l);
   }
}


void ROp(int *status,int *ifail)
{
   int n,i0,i1;
   int min_stat,sum_stat;
   spinor_dble **wsd;
   hmc_parms_t hmc;

   *status=0;
   *ifail=0;
   hmc=hmc_parms();
   
   error(query_blk_flags(HMC_BLOCKS,SWB_UP2DATE,hmc.nb)!=1,1,"ROp [bndop.c]",
         "The swb arrays on the HMC_BLOCKS grid are not up-to-date");

   n=query_blk_flags(HMC_BLOCKS,BCD_FLIPPED,0);
   error(((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)),1,
         "ROp [bndop.c]","Boundary conditions do not match data base");
   
   for (n=0;n<hmc.nb;n++)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0))
         blk_sw_term(n);

      if (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=1)
         *ifail|=invert_blk_swd(HMC_BLOCKS,n,ODD_PTS);
   }

   n=(*ifail);
   MPI_Reduce(&n,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD); 
   MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);
   
   if ((*ifail)!=0)
      return;

   error(check_blk_flags(HMC_BLOCKS,QHAT_DBLE,hmc.nb)!=0,1,
         "ROp [bndop.c]","Block fields are not in the proper condition");
   
   wsd=reserve_wsd(2);   
   i0=(wsd[0]-psd[0][0])/NSPIN;
   i1=(wsd[1]-psd[0][0])/NSPIN;   
   
   assign_chi2sd(i0);
   set_sd2zero(VOLUME,wsd[1]);
   
   DdOs(0,i0,i0);
   DOinv(1,i0,i0,status);
   DdOs(1,i0,i1);
   DOinv(0,i1,i1,status);

   sub_assign_PdOs(i0,i1);
   assign_sd2chi(i0);
   
   MPI_Reduce(status,&min_stat,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
   MPI_Reduce(status,&sum_stat,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);   
   MPI_Bcast(&min_stat,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&sum_stat,1,MPI_INT,0,MPI_COMM_WORLD);
   
   if (min_stat<0)
      *status=min_stat;
   else
      *status=sum_stat/(NPROC*hmc.nb);

   release_wsd();   
}
