
/*******************************************************************************
*
* File force1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the force F_1 that derives from the block pseudo-fermion
* action and the determinants of the SW-term
*
* The externally accessible functions are
*
*   void hmc_frf(int n,int *status,int *ifail)
*     Computes the force on the n'th block in the HMC_BLOCKS grid. On exit
*     status reports the  number of CG iterations that were required in this
*     calculation or a negative value if the program failed. The SW-term
*     is recomputed and inverted when needed and ifail is set to 0 or 1
*     depending on whether the inversions were safe or not
*
*   double hmc_acf(int n,int iphi,int idet,int *status,int *ifail)
*     Returns the pseudo-fermion action on the n'th block of the HMC_BLOCKS
*     grid (if iphi=1) plus the contribution of the determinants of the
*     SW-term to the block action (if idet=1). On exit status reports the 
*     number of BiCG iterations that were required in this calculation or
*     a negative value if the program failed. The SW-term is recomputed 
*     and inverted when needed and ifail is set to 0 or 1 depending on
*     whether the inversions were safe or not
*
* Notes:
*
* The programs in this file do not require any communications. Whether the
* SW-term on the blocks is in the proper condition is checked by querying
* the flags data base, and the proper action is then taken. Any errors
* are recorded by the error_loc() function
*
* The lattice parameters, maximal iteration numbers, ... are retrieved
* from the parameter data base using the utilities provided by the module
* flags/parms.c. Only the forces on the active links are calculated, and 
* the force field on inactive links is left untouched
*
* Note that the force field is a shared field. Moreover the same field is
* used for the various forces (gauge, block, block interaction)
*
*******************************************************************************/

#define FORCE1_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "linalg.h"
#include "linsolv.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "global.h"

#define MAX_LEVELS 8
#define BLK_LENGTH 8

static int cnt[MAX_LEVELS];
static double smx[MAX_LEVELS];
static block_t *bsv;


static void Op1(spinor *pk,spinor *pl)
{
   int k,l;

   k=(pk-(*bsv).s[0])/((*bsv).vol+1);
   l=(pl-(*bsv).s[0])/((*bsv).vol+1);
   
   Qhat_blk(bsv,k,l);
}


static void Op1_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l;

   k=(pk-(*bsv).sd[0])/((*bsv).vol+1);
   l=(pl-(*bsv).sd[0])/((*bsv).vol+1);
   
   Qhat_blk_dble(bsv,k,l);
}


static void third_vec(int volh,spinor_dble *sl,spinor_dble *phi,spinor_dble *st)
{
   spinor_dble *sm;

   sm=sl+volh;

   for (;sl<sm;sl++)
   {
      _vector_sub((*st).c1,(*phi).c1,(*sl).c1);
      _vector_sub((*st).c2,(*phi).c2,(*sl).c2);
      _vector_sub((*st).c3,(*phi).c3,(*sl).c3);      
      _vector_sub((*st).c4,(*phi).c4,(*sl).c4);      

      phi+=1;
      st+=1;
   }

   sm=sl+volh;

   for (;sl<sm;sl++)
   {
      _vector_mul((*st).c1,-1.0,(*sl).c1);
      _vector_mul((*st).c2,-1.0,(*sl).c2);
      _vector_mul((*st).c3,-1.0,(*sl).c3);
      _vector_mul((*st).c4,-1.0,(*sl).c4);      

      phi+=1;
      st+=1;
   }
}


void hmc_frf(int n,int *status,int *ifail)
{
   int nb,isw;
   spinor_dble *rb,*sb,*tb;
   hmc_t *hflds;
   hmc_parms_t hmc;

   *ifail=0;   
   bsv=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((hflds==NULL)||(n<0)||(n>=nb),1,"hmc_frf [force1.c]",
                 "Improper block grid, block fields or block number"))
   {
      *status=0;
      *ifail=0;
      return;
   }
   
   if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
       (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0))
      blk_sw_term(n);   

   if (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=1)
   {
      if (invert_blk_swd(HMC_BLOCKS,n,ODD_PTS))
      {
         *status=0;
         *ifail=1;
         return;
      }
   }

   assign_udb2ub(HMC_BLOCKS,n);
   assign_swdb2swb(HMC_BLOCKS,n,NO_PTS);

   bsv+=n;
   hflds+=n;
   rb=(*bsv).sd[0];
   sb=(*bsv).sd[1];
   tb=(*bsv).sd[2];
   
   apply_sw_dble((*bsv).volh,(*bsv).swd,(*hflds).phi,sb);   

   hmc=hmc_parms();
   cgne((*bsv).volh,0,Op1,Op1_dble,(*bsv).s,(*bsv).sd+2,
        hmc.nmx_frf,hmc.res_frf,sb,rb,status);

   Qhat_blk_dble(bsv,0,1);
   Qoe_blk_dble(bsv,0,0);
   Qoe_blk_dble(bsv,1,1);

   apply_sw_dble((*bsv).volh,(*bsv).swd+(*bsv).vol,
                 rb+(*bsv).volh,rb+(*bsv).volh);   
   apply_sw_dble((*bsv).volh,(*bsv).swd+(*bsv).vol,
                 sb+(*bsv).volh,sb+(*bsv).volh);   
   
   third_vec((*bsv).volh,sb,(*hflds).phi,tb);

   if (sw_frc(n,0,2,1))
   {
      *ifail=1;
      return;
   }

   hop_frc(n,0,1);
}


static void Op2_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l;

   k=(pk-(*bsv).sd[0])/((*bsv).vol+1);
   l=(pl-(*bsv).sd[0])/((*bsv).vol+1);
   
   Qhat_blk_dble(bsv,k,l);
   mulg5_dble((*bsv).volh,pl);
}


static double act_det(int vol,pauli_dble *m)
{
   int n;
   double c,p,act;
   pauli_dble *mm,*mn;
   lat_parms_t lat;

   lat=lat_parms();
   c=pow(4.0+lat.m0,-6.0);

   for (n=0;n<MAX_LEVELS;n++)
   {
      cnt[n]=0;
      smx[n]=0.0;
   }
   
   mm=m+2*vol;

   for (;m<mm;)
   {
      mn=m+BLK_LENGTH;
      if (mn>mm)
         mn=mm;
      p=1.0;

      for (;m<mn;m++)
         p*=(c*det_pauli_dble(m));

      cnt[0]+=1;
      smx[0]-=log(fabs(p));

      for (n=1;(cnt[n-1]>=BLK_LENGTH)&&(n<MAX_LEVELS);n++)
      {
         cnt[n]+=1;
         smx[n]+=smx[n-1];

         cnt[n-1]=0;
         smx[n-1]=0.0;
      }
   }

   act=0.0;

   for (n=0;n<MAX_LEVELS;n++)
      act+=smx[n];   
   
   return act+act;
}


double hmc_acf(int n,int iphi,int idet,int *status,int *ifail)
{
   int nb,isw;
   double act;
   spinor_dble *rb,*sb;
   hmc_t *hflds;
   hmc_parms_t hmc;

   *status=0;
   *ifail=0;

   if ((iphi!=1)&&(idet!=1))
      return 0.0;
   
   bsv=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((hflds==NULL)||(n<0)||(n>=nb),1,"hmc_acf [force1.c]",
                 "Improper block grid, block fields or block number"))
      return 0.0;

   act=0.0;
   bsv+=n;
   hflds+=n;   

   if (idet==1)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0)||
          (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=0))
         blk_sw_term(n);   

      act+=act_det((*bsv).vol,(*bsv).swd);
   }

   if (iphi==1)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0))
         blk_sw_term(n);   

      if (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=1)
      {
         if (invert_blk_swd(HMC_BLOCKS,n,ODD_PTS))
         {
            *ifail=1;
            return 0.0;
         }
      }

      rb=(*bsv).sd[0];
      sb=(*bsv).sd[1];
   
      apply_sw_dble((*bsv).volh,(*bsv).swd,(*hflds).phi,sb);   
      mulg5_dble((*bsv).volh,sb);

      hmc=hmc_parms();   
      bicg((*bsv).volh,0,Op2_dble,(*bsv).sd+2,hmc.nmx_acf,hmc.res_acf,
           sb,rb,status);

      if ((*status)>=0)
         act+=norm_square_dble((*bsv).volh,0,rb);
   }
   
   return act; 
}

