
/*******************************************************************************
*
* File force2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the force F_2 that derives from the block-interaction part
* of the HMC action
*
* The externally accessible functions are
*
*   void hmc_gcr_vec(int k1,int k2,int *status,int *ifail)
*     Computes the global spinor fields
*
*       psi1=Q^(-1)*Q_dOmega^star*chi and  psi2=Q^(-1)*R_2^(-1)*chi
*
*     and assigns them to psd[k1] and psd[k2] respectively. On exit status
*     reports the average number of Krylov vectors that were required by
*     the SAP+GCR solver in this calculation, or a negative value if the
*     solver failed. The global swd field and the fields on the GCR_BLOCKS
*     grid are updated automatically when needed, and ifail is set 0 or 1
*     depending on whether the inversions of the SW-term (if any) were safe
*     or not
*
*   void hmc_gcr_frf(int n,int k1,int k2)
*     Computes the force on the n'th block in the HMC_BLOCKS grid, assuming
*     that the fields psi1 and psi2 defined above are stored in the global
*     arrays psd[k1] and psd[k2]
*
*   double hmc_gcr_acf(int k1,int *status,int *ifail)
*     Computes the action of the pseudo-fermion field chi, assuming that
*     the global array psd[k1] contains an approximation to the field psi1
*     defined above. On exit status reports the  number of Krylov vectors
*     that were required by the SAP+GCR solver in this calculation, or a
*     negative value if the program failed. The SW-term is recomputed and
*     inverted when needed and ifail is set to 0 or 1 depending on whether
*     the inversions (if any) were safe or not
*
* Notes:
*
* All programs in this file except for hmc_gcr_frf() must be called on all
* processes simultaneously. Whether the SW-term is in the proper condition
* is checked by querying the flags data base
*
* In all cases the gauge field that is used is the double-precision field
* on the blocks of the HMC_BLOCKS grid. The inactive link variables on the
* blocks must match the ones on the global lattice when psi1 and psi2 or
* the action are calculated
*
* The lattice parameters, maximal iteration numbers, ... are retrieved
* from the parameter data base using the utilities provided by the module
* flags/parms.c. Only the forces on the active links are calculated, and 
* the force field on the inactive links is left untouched
*
* The programs hmc_gcr_vec() an hmc_gcr_acf() require a workspace of 1 and 
* 2 double-precision spinor fields respectively (in addition to the workspace
* needed by the SAP+GCR solver). The workspace is reserved using the utilties
* in the module flags/wspace.c
*
* Note that the force field is a shared field. Moreover the same field is
* used for the various forces (gauge, block, block interaction)
*
*******************************************************************************/

#define FORCE2_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "linalg.h"
#include "linsolv.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "hmc.h"
#include "global.h"


void hmc_gcr_vec(int k1,int k2,int *status,int *ifail)
{
   int n,nb,isw,iwd,iprms[2];
   int s1,s2,i1,i2;
   spinor_dble **wsd;
   block_t *b,*bm;
   hmc_parms_t hmc;

   iprms[0]=k1;
   iprms[1]=k2;
   
   MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

   error((iprms[0]!=k1)||(iprms[1]!=k2),1,"hmc_gcr_vec [force2.c]",
         "Parameters k1,k2 are not global");  

   hmc=hmc_parms();   
   n=query_flags(BCD_FLIPPED);
   error(((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)),1,
         "hmc_gcr_vec [force2.c]","Boundary conditions do not match data base");
   
   wsd=reserve_wsd(1);
   iwd=(wsd[0]-psd[0][0])/NSPIN;

   error_root((k1<0)||(k1>=no_sd)||(k2<0)||(k2>=no_sd)||
              (k1==k2)||(k1==iwd)||(k2==iwd),1,"hmc_gcr_vec [force2.c]",
              "Improper parameters k1,k2");  

   assign_chi2sd(iwd);
   set_sd2zero(VOLUME,psd[k2][0]);
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,iwd,0);
      Qbnds_dble(b,0,0);
      sub_assign_sdbb2sd(b,0,k2);
   } 

   mulmg5_dble(VOLUME,psd[k2][0]);
   swap_active(HMC_BLOCKS);

   sap_gcr(hmc.nmx_gcr_frf,hmc.res_gcr_frf,k2,k1,&s1,&i1);

   if (i1==0)
   {
      sub_assign_PdOs(iwd,k1);
      mulg5_dble(VOLUME,wsd[0]);
      sap_gcr(hmc.nmx_gcr_frf,hmc.res_gcr_frf,iwd,k2,&s2,&i2);
   }
   else
   {
      s2=0;
      i2=0;
   }
   
   swap_active(HMC_BLOCKS);

   if (s1<0)
      *status=s1;
   else if (s2<0)
      *status=s2;
   else
      *status=(s1+s2)/2;

   *ifail=(i1|i2);
   release_wsd();
}


void hmc_gcr_frf(int n,int k1,int k2)
{
   int nb,isw;
   block_t *b;

   b=blk_list(HMC_BLOCKS,&nb,&isw);

   if (error_loc((b==NULL)||(n<0)||(n>=nb),1,"hmc_gcr_frf [force2.c]",
                 "Block grid is not allocated or block number out of range"))
      return;

   if (error_loc((k1<0)||(k1>=no_sd)||(k2<0)||(k2>=no_sd),1,
                 "hmc_gcr_frf [force2.c]","Improper arguments k1,k2"))
      return;

   b+=n;
   assign_sd2sdb(b,ALL_PTS,k1,0);
   assign_sd2sdb(b,ALL_PTS,k2,1);
   
   if (sw_frc(n,0,1,0)==0)
      hop_frc(n,0,1);
}


double hmc_gcr_acf(int k1,int *status,int *ifail)
{
   int n,nb,isw,iwd0,iwd1,iprms[1];
   double res0,res1,res,act;
   complex_dble z;
   spinor_dble **wsd;
   block_t *b,*bm;
   hmc_parms_t hmc;

   iprms[0]=k1;
   
   MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);

   error(iprms[0]!=k1,1,"hmc_gcr_acf [force2.c]","Parameter k1 is not global");

   hmc=hmc_parms();   
   n=query_flags(BCD_FLIPPED);
   error(((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)),1,
         "hmc_gcr_acf [force2.c]","Boundary conditions do not match data base");
   
   wsd=reserve_wsd(2);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;
   iwd1=(wsd[1]-psd[0][0])/NSPIN;
   
   error_root((k1<0)||(k1>=no_sd)||(k1==iwd0)||(k1==iwd1),1,
              "hmc_gcr_acf [force2.c]","Improper parameter k1");
   
   assign_chi2sd(iwd0);
   set_sd2zero(VOLUME,wsd[1]);

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   error(b==NULL,1,"hmc_gcr_acf [force2.c]","Block grid is not allocated");
   bm=b+nb;

   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,iwd0,0);
      Qbnds_dble(b,0,0);
      sub_assign_sdbb2sd(b,0,iwd1);
   } 
   
   res0=norm_square_dble(VOLUME,1,wsd[1]);
   swap_active(HMC_BLOCKS);

   if ((query_flags(SWD_UP2DATE)!=1)||
       (query_flags(SWD_E_INVERTED)!=0)||(query_flags(SWD_O_INVERTED)!=0))
      sw_term();
   
   Qnohat_dble(k1,iwd0);
   z.re=1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,wsd[0],wsd[1],z);
   res1=norm_square_dble(VOLUME,1,wsd[0]);
   res=hmc.res_gcr_acf;
   
   if (res1<res0)
   {
      if (res1>(res0*res*res))
      {
         res*=sqrt(res0/res1);
         mulmg5_dble(VOLUME,wsd[0]);
         sap_gcr(hmc.nmx_gcr_acf,res,iwd0,iwd1,status,ifail);         

         if ((*ifail)==0)
         {
            mulc_spinor_add_dble(VOLUME,wsd[1],psd[k1][0],z);
            assign_chi2sd(iwd0);   
            sub_assign_PdOs(iwd0,iwd1);
            act=norm_square_dble(VOLUME,1,wsd[0]);
         }
         else
            act=0.0;
      }
      else
      {
         *status=0;
         *ifail=0;
         assign_chi2sd(iwd0);   
         sub_assign_PdOs(iwd0,k1);
         act=norm_square_dble(VOLUME,1,wsd[0]);
      }
   }
   else
   {
      mulmg5_dble(VOLUME,wsd[1]);
      sap_gcr(hmc.nmx_gcr_acf,res,iwd1,iwd0,status,ifail);         
      
      if ((*ifail)==0)
      {
         assign_chi2sd(iwd1);   
         sub_assign_PdOs(iwd1,iwd0);
         act=norm_square_dble(VOLUME,1,wsd[1]);
      }
      else
         act=0.0;
   }

   swap_active(HMC_BLOCKS);
   release_wsd();

   return act;
}

