
/*******************************************************************************
*
* File liealg.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Basic functions for fields with values in the Lie algebra of SU(3)
*
* The externally accessible functions are
*
*   void ExpX(double t,su3_alg_dble *X,su3_dble *u)
*     Replaces u by E(t*X)*u where E(X) is an approximation to the matrix
*     exponential function. E(X) is in SU(3) and satisfies E(X)*E(-X)=1
*     to machine precision as well as E(t*X)=1+t*X+(1/2)*t^2*X^2+O(t^3)
*
*   void random_alg(int n,su3_alg_dble *X)
*     Initializes the Lie algebra elements X[0],..,X[n-1] to random 
*     values with distribution proportional to exp{tr[X^2]}
*
*   double norm_square_alg(int n,su3_alg_dble *X)
*     Returns sum_{k=0}^{n-1}(-2)*tr{X[k]^2}, i.e. the norm squared of 
*     the field X[]
*
*   void set_alg2zero(int n,su3_alg_dble *X)
*     Sets the array elements X[0],..,X[n-1] to zero
*
* Notes:
*
* Lie algebra elements X are traceless antihermitian 3x3 matrices that
* are represented by structures with real elements x1,...,x8 through
*
*  X_11=i*(x1+x2), X_22=i*(x2-2*x1), X_33=i*(x1-2*x2), 
*
*  X_12=x3+i*x4, X_13=x5+i*x6, X_23=x7+i*x8
*
* All programs in this file are for double-precision variables. The
* function E(X) is constructed using a product of SU(2) matrices
*
*******************************************************************************/

#define LIEALG_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "start.h"
#include "random.h"
#include "global.h"

#define MAX_LEVELS 8
#define BLK_LENGTH 8

static int cnt[MAX_LEVELS];
static double smx[MAX_LEVELS];
static double r1,s1[4],s2[4],s3[4];
static double c1=0.0,c2,c3,rb[8];
static complex_dble z1,z2;
static const su3_alg_dble X0={0.0};


static void su2_rotate(double *s,su3_vector_dble *v1,su3_vector_dble *v2)
{
   z1.re=
      s[0]*(*v1).c1.re-s[1]*(*v2).c1.im+s[2]*(*v2).c1.re-s[3]*(*v1).c1.im;
   z1.im=
      s[0]*(*v1).c1.im+s[1]*(*v2).c1.re+s[2]*(*v2).c1.im+s[3]*(*v1).c1.re;
   z2.re=
      s[0]*(*v2).c1.re-s[1]*(*v1).c1.im-s[2]*(*v1).c1.re+s[3]*(*v2).c1.im;
   z2.im=
      s[0]*(*v2).c1.im+s[1]*(*v1).c1.re-s[2]*(*v1).c1.im-s[3]*(*v2).c1.re;
   (*v1).c1=z1;
   (*v2).c1=z2;

   z1.re=
      s[0]*(*v1).c2.re-s[1]*(*v2).c2.im+s[2]*(*v2).c2.re-s[3]*(*v1).c2.im;
   z1.im=
      s[0]*(*v1).c2.im+s[1]*(*v2).c2.re+s[2]*(*v2).c2.im+s[3]*(*v1).c2.re;
   z2.re=
      s[0]*(*v2).c2.re-s[1]*(*v1).c2.im-s[2]*(*v1).c2.re+s[3]*(*v2).c2.im;
   z2.im=
      s[0]*(*v2).c2.im+s[1]*(*v1).c2.re-s[2]*(*v1).c2.im-s[3]*(*v2).c2.re;
   (*v1).c2=z1;
   (*v2).c2=z2;

   z1.re=
      s[0]*(*v1).c3.re-s[1]*(*v2).c3.im+s[2]*(*v2).c3.re-s[3]*(*v1).c3.im;
   z1.im=
      s[0]*(*v1).c3.im+s[1]*(*v2).c3.re+s[2]*(*v2).c3.im+s[3]*(*v1).c3.re;
   z2.re=
      s[0]*(*v2).c3.re-s[1]*(*v1).c3.im-s[2]*(*v1).c3.re+s[3]*(*v2).c3.im;
   z2.im=
      s[0]*(*v2).c3.im+s[1]*(*v1).c3.re-s[2]*(*v1).c3.im-s[3]*(*v2).c3.re;
   (*v1).c3=z1;
   (*v2).c3=z2;   
}


void ExpX(double t,su3_alg_dble *X,su3_dble *u)
{
   su3_vector_dble *v1,*v2,*v3;

   s1[1]=t*(*X).c4;
   s1[2]=t*(*X).c3;
   s1[3]=t*(*X).c1;

   r1=s1[1]*s1[1]+s1[2]*s1[2]+s1[3]*s1[3];   
   r1=1.0/(2.0+0.125*r1);
   s1[0]=4.0*r1-1.0;
   s1[1]*=r1;
   s1[2]*=r1;
   s1[3]*=r1;   
   
   s2[1]=t*(*X).c6;
   s2[2]=t*(*X).c5;
   s2[3]=t*(*X).c2;   

   r1=s2[1]*s2[1]+s2[2]*s2[2]+s2[3]*s2[3];   
   r1=1.0/(2.0+0.125*r1);
   s2[0]=4.0*r1-1.0;
   s2[1]*=r1;
   s2[2]*=r1;
   s2[3]*=r1;  
   
   s3[1]=t*(*X).c8;
   s3[2]=t*(*X).c7;
   s3[3]=t*((*X).c2-(*X).c1);
   
   r1=s3[1]*s3[1]+s3[2]*s3[2]+s3[3]*s3[3];   
   r1=1.0/(1.0+0.25*r1);
   s3[0]=2.0*r1-1.0;
   s3[1]*=r1;
   s3[2]*=r1;
   s3[3]*=r1;

   v1=(su3_vector_dble*)(u);
   v2=v1+1;
   v3=v1+2;
   
   su2_rotate(s1,v1,v2);
   su2_rotate(s2,v1,v3);   
   su2_rotate(s3,v2,v3);
   su2_rotate(s2,v1,v3);
   su2_rotate(s1,v1,v2);
}


void random_alg(int n,su3_alg_dble *X)
{
   su3_alg_dble *rx;

   if (c1==0.0)
   {
      c1=(sqrt(3.0)+1.0)/6.0;
      c2=(sqrt(3.0)-1.0)/6.0;
      c3=1/sqrt(2.0);
   }
   
   for (rx=X;rx<(X+n);rx++)
   {
      gauss_dble(rb,8);

      (*rx).c1=c1*rb[0]+c2*rb[1];
      (*rx).c2=c1*rb[1]+c2*rb[0];      

      (*rx).c3=c3*rb[2];
      (*rx).c4=c3*rb[3];
      (*rx).c5=c3*rb[4];
      (*rx).c6=c3*rb[5];
      (*rx).c7=c3*rb[6];
      (*rx).c8=c3*rb[7];
   }
}


double norm_square_alg(int n,su3_alg_dble *X)
{
   su3_alg_dble *rx,*rxm,*rxt;

   rx=X;
   rxm=rx;
   rxt=rx+n;

   for (n=0;n<MAX_LEVELS;n++)
   {
      cnt[n]=0;
      smx[n]=0.0;
   }
   
   for (;rxm<rxt;)
   {
      rxm+=BLK_LENGTH;
      if (rxm>rxt)
         rxm=rxt;
      r1=0.0;

      for (;rx<rxm;)
      {
         r1+=3.0*((*rx).c1*(*rx).c1+(*rx).c2*(*rx).c2-(*rx).c1*(*rx).c2)+
                  (*rx).c3*(*rx).c3+(*rx).c4*(*rx).c4+(*rx).c5*(*rx).c5+
                  (*rx).c6*(*rx).c6+(*rx).c7*(*rx).c7+(*rx).c8*(*rx).c8;
         rx+=1;
      }

      cnt[0]+=1;
      smx[0]+=r1;

      for (n=1;(cnt[n-1]>=BLK_LENGTH)&&(n<MAX_LEVELS);n++)
      {
         cnt[n]+=1;
         smx[n]+=smx[n-1];

         cnt[n-1]=0;
         smx[n-1]=0.0;
      }
   }

   r1=0.0;

   for (n=0;n<MAX_LEVELS;n++)
      r1+=smx[n];

   return 4.0*r1;
}


void set_alg2zero(int n,su3_alg_dble *X)
{
   su3_alg_dble *rx;

   for (rx=X;rx<(X+n);rx++)
      *rx=X0;
}
