
/*******************************************************************************
*
* File lpfrog.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Leap-frog integration of the HMD equations
*
* The externally accessible functions are
*
*   double start_hmd(int *stat_rop,int *ifail)
*     Initialization of the momentum fields, the pseudo-fermion fields and
*     the gauge fields on the blocks of the HMC_BLOCKS grid. The program
*     returns the active part of the total action of these fields. On exit
*     stat_rop reports the average number of BiCGstab iterations that were
*     required when calculating R_2*chi, or a negative value if the program
*     failed. The SW-term is recomputed and inverted when needed and ifail
*     is set to 0 or 1 depending on whether the inversions (if any) were
*     safe or not
*
*   double run_hmd(int *stat_acf,int *stat_gcr_acf,
*                  int *stat_frf,int *stat_gcr_frf,int *ifail)
*     This program evolves the gauge field and the momenta on the blocks of
*     the HMC_BLOCKS grid according to the parameters set by set_hmc_parms()
*     and returns the active part of the action of the final configuration.
*     On exit the parameters stat_* report the average solver iteration
*     numbers per application of the BiCGstab, SAP+GCR and CG solvers which
*     were required along the trajectory (or a negative value if the solver
*     failed). The SW-term is recomputed and inverted when needed and ifail
*     is set to 0 or 1 depending on whether the inversions (if any) were
*     safe or not
*
*   int accept_step(double act0,double act1)
*     Returns 1 with probability min{1,exp(act0-act1)} and 0 otherwise.
*     If 1 is returned, the active link variables on the HMC_BLOCKS grid
*     are swapped with those in the global field. On exit the time-like
*     link variables on the global lattice at time NPROC0*L0-1 are left
*     in the unflipped state and the field is renormalized to SU(3) if 
*     it was changed
*
* Notes:
*
* The gauge field that is evolved in the course of the leap-frog integration
* is the double-precision field on the blocks. The global double-precision
* gauge field is left intact until the evolved configuration is accepted
*
* It is assumed that the parameters of the algorithm have been set by calling
* set_lat_parms(), set_gcr_parms() and set_hmc_parms() [flags/parms.c]. Apart
* from the fields used by the calling programs, 2*nkv+1 single-precision and
* 4 double-precision spinor fields must be allocated (where nkv denotes the
* number of Krylov vectors that was entered when set_gcr_parms() was called)
*
* If any of the status variables reports a negative value, or if the argument
* *ifail is returned with a non-zero value, the programs have not completed
* correctly and the calling program should invoke an error handling procedure
* in this case
*
*******************************************************************************/

#define LPFROG_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "random.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "linalg.h"
#include "linsolv.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "hmc.h"
#include "global.h"

static int iwd0,iwd1;

#if (defined SSE2)
#include "sse2.h"
static pauli_dble minv __attribute__ ((aligned (16)));
#else
static pauli_dble minv;
#endif


static int apply_Qee_inv(int vol,pauli_dble *m,spinor_dble *s,spinor_dble *r)
{
   int ifail;
   weyl_dble *rs,*rr,*rsm;

   ifail=0;
   rs=(weyl_dble*)(s);
   rr=(weyl_dble*)(r);
   rsm=rs+2*vol;

   for (;rs<rsm;rs++)
   {
      ifail|=inv_pauli_dble(m,&minv);
      mul_pauli_dble(&minv,rs,rr);
      m+=1;
      rr+=1;
   }

   return ifail;
}


double start_hmd(int *stat_rop,int *ifail)
{
   int n,nb,nbh,isw,flag,dmy0,dmy1;
   double fact,blk_act,act0;
   block_t *b;
   hmc_t *hflds;
   lat_parms_t lat;
   gcr_parms_t gcr;
   hmc_parms_t hmc;

   lat=lat_parms();
   gcr=gcr_parms();
   hmc=hmc_parms();
   
   error_root((gcr.nb==0)||(hmc.nb==0),1,"start_hmd [lpfrog.c]",
              "Block grids are not allocated");

   n=query_flags(BCD_FLIPPED);
   if (((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)))
      flipbcd();

   assign_ud2udbgr(HMC_BLOCKS);
   swb_term();
   random_chi();

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();
   nbh=nb/2;
   fact=lat.beta/6.0;
   blk_act=0.0;
   flag=0;
   
   for (n=0;n<nb;n++)
   {   
      random_alg((*hflds).nlks,(*hflds).mom);
      blk_act+=0.5*norm_square_alg((*hflds).nlks,(*hflds).mom);

      blk_act+=fact*hmc_acg(n);
      blk_act+=hmc_acf(n,0,1,&dmy0,&dmy1);      
      
      random_sd((*b).volh,(*b).sd[0],1.0);
      blk_act+=norm_square_dble((*b).volh,0,(*b).sd[0]);

      flag|=invert_blk_swd(HMC_BLOCKS,n,ODD_PTS);
      Qhat_blk_dble(b,0,1);
      flag|=apply_Qee_inv((*b).volh,(*b).swd,(*b).sd[1],(*hflds).phi);

      if (((isw==0)&&(n<nbh))||((isw==1)&&(n>=nbh)))
         blk_act+=norm_square_dble((*hflds).nbnd,0,(*hflds).chi);

      b+=1;
      hflds+=1;
   }
   
   ROp(stat_rop,ifail);

   flag|=(*ifail);
   MPI_Reduce(&flag,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);

   MPI_Reduce(&blk_act,&act0,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(&act0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   error_chk();   
   
   return act0;
}


static void updt_mom(int n,double eps)
{
   su3_alg_dble *frc,*frm,*mom;
   hmc_t *hflds;

   hflds=hmcflds()+n;
   mom=(*hflds).mom;
   frc=(*hflds).frc;
   frm=frc+(*hflds).nlks;

   for (;frc<frm;frc++)
   {
      (*mom).c1-=eps*(*frc).c1;
      (*mom).c2-=eps*(*frc).c2;      
      (*mom).c3-=eps*(*frc).c3;
      (*mom).c4-=eps*(*frc).c4;
      (*mom).c5-=eps*(*frc).c5;
      (*mom).c6-=eps*(*frc).c6;      
      (*mom).c7-=eps*(*frc).c7;
      (*mom).c8-=eps*(*frc).c8;

      mom+=1;
   }
}


static void T_U(int n,double eps)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub,*u;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   mom=(*hflds).mom;
   
   for (;lk<lkm;lk++)
   {
      u=ub+(*lk);
      ExpX(eps,mom,u);
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static void T_G(int n,double eps)
{
   hmc_frg(n);
   updt_mom(n,eps);
}


static void T_F(int n,double eps,int *status,int *ifail)
{
   int stat,flag;

   hmc_frf(n,&stat,&flag);
   
   if (stat<0)
      (*status)=stat;
   else
      (*status)+=stat;

   (*ifail)|=flag;

   if (((*status)<0)||((*ifail)!=0))
      return;

   updt_mom(n,eps);
}


static void T_GCR(double eps,int *status,int *ifail)
{
   int n,nb,isw,stat,flag;

   hmc_gcr_vec(iwd0,iwd1,&stat,&flag);

   if (stat<0)
      *status=stat;
   else
      *status+=stat;

   (*ifail)|=flag;

   if (((*status)<0)||((*ifail)!=0))
      return;
   
   blk_list(HMC_BLOCKS,&nb,&isw);
   
   for (n=0;n<nb;n++)
   {
      hmc_gcr_frf(n,iwd0,iwd1);
      updt_mom(n,eps);
   }
}


double run_hmd(int *stat_acf,int *stat_gcr_acf,
               int *stat_frf,int *stat_gcr_frf,int *ifail)
{
   int n,nb;
   int n0,n1,n2;
   int k0,k1,k2;
   int stat,flag,stat_min,stat_sum;
   double fact,eps0,eps1,eps2,blk_act,act0,act1;
   spinor_dble **wsd;
   hmc_t *hflds;
   lat_parms_t lat;
   hmc_parms_t hmc;

   wsd=reserve_wsd(2);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;
   iwd1=(wsd[1]-psd[0][0])/NSPIN;   

   lat=lat_parms();
   hmc=hmc_parms();

   nb=hmc.nb;
   n0=hmc.N0;
   n1=hmc.N1;
   n2=hmc.N2;

   fact=lat.beta/6.0;   
   eps2=hmc.tau/(double)(n2);
   eps1=eps2/(double)(n1);   
   eps0=eps1/(double)(n0);

   *stat_acf=0;
   *stat_gcr_acf=0;
   *stat_frf=0;
   *stat_gcr_frf=0;
   *ifail=0;

   for (k2=1;k2<=n2;k2++)
   {
      if (k2==1)
         T_GCR(0.5*eps2,stat_gcr_frf,ifail);
      else
      {
         update_swdbgr2swd();
         T_GCR(eps2,stat_gcr_frf,ifail);
      }

      if (((*stat_gcr_frf)<0)||((*ifail)!=0))
         break;

      stat=0;
      flag=0;
      
      for (n=0;n<nb;n++)
      {
         for (k1=1;k1<=n1;k1++)
         {
            if ((k2==1)&&(k1==1))
               T_F(n,0.5*eps1,&stat,&flag);
            else
               T_F(n,eps1,&stat,&flag);

            if ((stat<0)||(flag!=0))
               break;
            
            for (k0=1;k0<=n0;k0++)
            {
               if ((k2==1)&&(k1==1)&&(k0==1))
                  T_G(n,0.5*fact*eps0);
               else
                  T_G(n,fact*eps0); 

               T_U(n,eps0);

               if ((k2==n2)&&(k1==n1)&&(k0==n0))
                  T_G(n,0.5*fact*eps0); 
            }
         
            if ((k2==n2)&&(k1==n1))
               T_F(n,0.5*eps1,&stat,&flag);

            if ((stat<0)||(flag!=0))
               break;
         }

         if ((stat<0)||(flag!=0))
            break;
      }

      MPI_Reduce(&stat,&stat_min,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
      MPI_Reduce(&stat,&stat_sum,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Reduce(&flag,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&stat_min,1,MPI_INT,0,MPI_COMM_WORLD);      
      MPI_Bcast(&stat_sum,1,MPI_INT,0,MPI_COMM_WORLD);            
      MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);
      
      if (stat_min<0)
         (*stat_frf)=stat_min;
      else
         (*stat_frf)+=stat_sum;

      if (((*stat_frf)<0)||((*ifail)!=0))
         break;      
      
      if (k2==n2)
      {
         update_swdbgr2swd();
         T_GCR(0.5*eps2,stat_gcr_frf,ifail);
      }
      
      if (((*stat_gcr_frf)<0)||((*ifail)!=0))
         break;
   }

   if ((*stat_frf)>=0)
      (*stat_frf)/=(NPROC*nb*(n1*n2+1));
   
   if ((*stat_gcr_frf)>=0)
      (*stat_gcr_frf)/=(n2+1);

   release_wsd();
   
   if (((*stat_frf)<0)||((*stat_gcr_frf)<0)||((*ifail)!=0))
      return 0.0;

   reserve_wsd(1);
   act0=hmc_gcr_acf(iwd0,stat_gcr_acf,ifail);
   release_wsd();
   
   if (((*stat_gcr_acf)<0)||((*ifail)!=0))
      return 0.0;

   hflds=hmcflds();
   stat=0;
   flag=0;
   blk_act=0.0;
   
   for (n=0;n<nb;n++)
   {
      blk_act+=0.5*norm_square_alg((*hflds).nlks,(*hflds).mom);
      blk_act+=fact*hmc_acg(n);
      blk_act+=hmc_acf(n,1,1,stat_acf,ifail);

      if ((*stat_acf)>=0)
         stat+=(*stat_acf);
      else
         stat=(*stat_acf);

      flag|=(*ifail);

      if ((stat<0)||(flag!=0))
         break;
      
      hflds+=1;
   }

   MPI_Reduce(&stat,&stat_min,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
   MPI_Reduce(&stat,&stat_sum,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Reduce(&flag,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&stat_min,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(&stat_sum,1,MPI_INT,0,MPI_COMM_WORLD);            
   MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);
      
   if (stat_min<0)
      (*stat_acf)=stat_min;
   else
      (*stat_acf)=stat_sum/(NPROC*nb);

   if (((*stat_acf)<0)||((*ifail)!=0))
      return 0.0;
   
   MPI_Reduce(&blk_act,&act1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(&act1,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   error_chk();   

   return act0+act1;
}


int accept_step(double act0,double act1)
{
   int my_rank,iac;
   double dprms[2],r;

   dprms[0]=act0;
   dprms[1]=act1;

   MPI_Bcast(dprms,2,MPI_DOUBLE,0,MPI_COMM_WORLD);   

   error((dprms[0]!=act0)||(dprms[1]!=act1),1,"accept_step [lpfrog.c]",
         "Parameters are not global");
   
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      ranlxd(&r,1);

      if (r<=exp(act0-act1))
         iac=1;
      else
         iac=0;
   }
   
   MPI_Bcast(&iac,1,MPI_INT,0,MPI_COMM_WORLD);   

   if (iac==1)
      swap_active(HMC_BLOCKS);

   if (query_flags(BCD_FLIPPED)!=0)
      flipbcd();

   if (iac==1)
      renormalize_ud();
   
   return iac;
}

