
/*******************************************************************************
*
* File archive.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs to read and write field configurations
*
* The externally accessible functions are
*
*   int copy_file(char *in,char *out)
*     Copies the file "in" to the file "out" in binary mode. The return
*     value is 0 if no I/O error is detected and 1 otherwise, in which
*     case the error can also be detected by the error_chk() [utils.c]
*     function
*
*   void write_cnfg(char *out)
*     Writes the lattice sizes, the processor grid, the rank of the
*     calling process, the state of the random number generator and the
*     local double-precision gauge field to the file "out"
*
*   void read_cnfg(char *in)
*     Reads the data previously written by the program write_cnfg from
*     the file "in" and resets the random number generator and the local
*     double-precision gauge field accordingly
*
*   void export_cnfg(char *out)
*     Writes the lattice sizes and the global double-precision gauge field to
*     the file "out" from process 0 in the universal format specified below
*
*   void import_cnfg(char *in)
*     Reads the global double-precision gauge field from the file "in" and
*     extends the field periodically if needed. The file is read by process
*     0 only and it is assumed that the link variables are stored on the
*     file in the universal format
*
* Notes:
*
* The program copy_file does not involve any communications and can be called
* locally. All other programs must be called simultaneously on all processes
*
* The program export_cnfg first writes the lattice sizes and the average of
* the plaquette Re tr{U_p} to the output file. Then follow the 8 link variables
* in the directions +0,-0,...,+3,-3 at the first odd point, the second odd
* point, and so on. The order of the point (x0,x1,x2,x3) with coordinates in
* the range 0<=x0<N0,...,0<=x3<N3 is determined by the index
*
*   ix=x3+N3*x2+N2*N3*x1+N1*N2*N3*x0
*
* where N0,N1,N2,N3 are the lattice sizes
*
*******************************************************************************/

#define ARCHIVE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "flags.h"
#include "update.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int ns,nd,*state=NULL;
static su3_dble *ubuf=NULL,*vbuf;


static void alloc_state(void)
{
   int n;

   ns=rlxs_size();
   nd=rlxd_size();

   if (ns<nd)
      n=nd;
   else
      n=ns;

   state=amalloc(n*sizeof(int),3);
   error(state==NULL,1,"alloc_state [archive.c]",
         "Unable to allocate auxiliary array");
}


static void alloc_ubuf(int my_rank)
{
   if (my_rank==0)
   {
      ubuf=amalloc(4*(L3+N3)*sizeof(su3_dble),4);
      vbuf=ubuf+4*L3;
   }
   else
      ubuf=amalloc(4*L3*sizeof(su3_dble),4);

   error(ubuf==NULL,1,"alloc_ubuf [archive.c]",
         "Unable to allocate auxiliary array");
}


static void get_links(int iy)
{
   int y3,iz,mu;
   su3_dble *u,*v;

   v=ubuf;
   iy*=L3;

   if (ipt[iy]<(VOLUME/2))
      iy+=1;

   for (y3=0;y3<L3;y3+=2)
   {
      iz=ipt[iy+y3];
      u=pud[iz][0];

      for (mu=0;mu<8;mu++)
      {
         *v=*u;
         v+=1;
         u+=1;
      }
   }
}


static void set_links(int iy)
{
   int y3,iz,mu;
   su3_dble *u,*v;

   v=ubuf;
   iy*=L3;

   if (ipt[iy]<(VOLUME/2))
      iy+=1;

   for (y3=0;y3<L3;y3+=2)
   {
      iz=ipt[iy+y3];
      u=pud[iz][0];

      for (mu=0;mu<8;mu++)
      {
         *u=*v;
         v+=1;
         u+=1;
      }
   }
}


int copy_file(char *in,char *out)
{
   int c;
   FILE *fin,*fout;

   fin=fopen(in,"rb");

   if (fin==NULL)
   {
      error_loc(1,1,"copy_file [archive.c]","Unable to open input file");
      return 1;
   }

   fout=fopen(out,"wb");

   if (fout==NULL)
   {
      error_loc(1,1,"copy_file [archive.c]","Unable to open output file");
      fclose(fin);
      return 1;
   }

   c=getc(fin);

   while (feof(fin)==0)
   {
      putc(c,fout);
      c=getc(fin);
   }

   if ((ferror(fin)==0)&&(ferror(fout)==0))
   {
      fclose(fin);
      fclose(fout);
      return 0;
   }
   else
   {
      error_loc(1,1,"copy_file [archive.c]","Read or write error");
      fclose(fin);
      fclose(fout);
      return 1;
   }
}


void write_cnfg(char *out)
{
   int ldat[9];
   FILE *fout;

   if (state==NULL)
      alloc_state();

   fout=fopen(out,"wb");
   error_loc(fout==NULL,1,"write_cnfg [archive.c]",
             "Unable to open output file");
   error_chk();

   ldat[0]=NPROC0;
   ldat[1]=NPROC1;
   ldat[2]=NPROC2;
   ldat[3]=NPROC3;

   ldat[4]=L0;
   ldat[5]=L1;
   ldat[6]=L2;
   ldat[7]=L3;

   MPI_Comm_rank(MPI_COMM_WORLD,ldat+8);

   fwrite(ldat,sizeof(int),9,fout);
   rlxs_get(state);
   fwrite(state,sizeof(int),ns,fout);
   rlxd_get(state);
   fwrite(state,sizeof(int),nd,fout);
   fwrite(pud[VOLUME/2][0],sizeof(su3_dble),4*VOLUME,fout);

   error_loc(ferror(fout)!=0,1,"write_cnfg [archive.c]","Write error");
   error_chk();
   fclose(fout);
}


void read_cnfg(char *in)
{
   int n,ldat[9];
   FILE *fin;

   if (state==NULL)
      alloc_state();

   error(pud[VOLUME/2][0]==NULL,1,"read_cnfg [archive.c]",
         "Double-precision gauge field is not allocated");

   fin=fopen(in,"rb");
   error_loc(fin==NULL,1,"read_cnfg [archive.c]",
             "Unable to open input file");
   error_chk();

   fread(ldat,sizeof(int),9,fin);
   MPI_Comm_rank(MPI_COMM_WORLD,&n);

   error((ldat[0]!=NPROC0)||(ldat[1]!=NPROC1)||
         (ldat[2]!=NPROC2)||(ldat[3]!=NPROC3)||
         (ldat[4]!=L0)||(ldat[5]!=L1)||(ldat[6]!=L2)||(ldat[7]!=L3)||
         (ldat[8]!=n),1,"read_cnfg [archive.c]","Unexpected lattice data");

   fread(state,sizeof(int),ns,fin);
   rlxs_reset(state);
   fread(state,sizeof(int),nd,fin);
   rlxd_reset(state);
   fread(pud[VOLUME/2][0],sizeof(su3_dble),4*VOLUME,fin);

   error_loc(ferror(fin)!=0,1,"read_cnfg [archive.c]","Read error");
   error_chk();
   fclose(fin);
   set_flags(NEW_UD);
}


void export_cnfg(char *out)
{
   int my_rank,np[4],n,tag;
   int x0,x1,x2,x3,y0,y1,y2,ix,iy;
   double plaq;
   MPI_Status stat;
   FILE *fout=NULL;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (ubuf==NULL)
      alloc_ubuf(my_rank);

   plaq=plaq_sum_dble()/(double)(6*NPROC*VOLUME);

   if (my_rank==0)
   {
      fout=fopen(out,"wb");
      error_root(fout==NULL,1,"export_cnfg [archive.c]",
                 "Unable to open output file");

      np[0]=N0;
      np[1]=N1;
      np[2]=N2;
      np[3]=N3;

      fwrite(np,sizeof(int),4,fout);
      fwrite(&plaq,sizeof(double),1,fout);
   }

   for (ix=0;ix<(N0*N1*N2);ix++)
   {
      x0=(ix/(N1*N2));
      x1=(ix/N2)%N1;
      x2=ix%N2;

      y0=x0%L0;
      y1=x1%L1;
      y2=x2%L2;
      iy=y2+L2*y1+L1*L2*y0;

      np[0]=x0/L0;
      np[1]=x1/L1;
      np[2]=x2/L2;

      for (x3=0;x3<N3;x3+=L3)
      {
         np[3]=x3/L3;
         n=ipr_global(np);
         if (my_rank==n)
            get_links(iy);
         MPI_Barrier(MPI_COMM_WORLD);

         if (n>0)
         {
            tag=mpi_tag();

            if (my_rank==n)
               MPI_Send((double*)(ubuf),4*L3*18,MPI_DOUBLE,0,
                        tag,MPI_COMM_WORLD);

            if (my_rank==0)
               MPI_Recv((double*)(ubuf),4*L3*18,MPI_DOUBLE,n,
                        tag,MPI_COMM_WORLD,&stat);
         }

         if (my_rank==0)
            fwrite(ubuf,sizeof(su3_dble),4*L3,fout);
      }
   }

   if (my_rank==0)
   {
      error_root(ferror(fout)!=0,1,
                 "export_cnfg [archive.c]","Write error");
      fclose(fout);
   }
}


void import_cnfg(char *in)
{
   int my_rank,np[4],n,tag;
   int k,l,x0,x1,x2,y0,y1,y2,y3,c0,c1,c2,ix,iy,ic;
   int n0,n1,n2,n3,nc0,nc1,nc2,nc3;
   double plaq0,plaq1;
   MPI_Status stat;
   FILE *fin=NULL;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (ubuf==NULL)
      alloc_ubuf(my_rank);

   error(pud[VOLUME/2][0]==NULL,1,"import_cnfg [archive.c]",
         "Double-precision gauge field is not allocated");

   if (my_rank==0)
   {
      fin=fopen(in,"rb");
      error_root(fin==NULL,1,"import_cnfg [archive.c]",
                 "Unable to open input file");

      fread(np,sizeof(int),4,fin);
      fread(&plaq0,sizeof(double),1,fin);

      error_root(((N0%np[0])!=0)||((N1%np[1])!=0)||
                 ((N2%np[2])!=0)||((N3%np[3])!=0),1,
                 "import_cnfg [archive.c]","Incompatible lattice sizes");
   }

   MPI_Bcast(np,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&plaq0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   n0=np[0];
   n1=np[1];
   n2=np[2];
   n3=np[3];

   nc0=N0/n0;
   nc1=N1/n1;
   nc2=N2/n2;
   nc3=N3/n3;

   for (ix=0;ix<(n0*n1*n2);ix++)
   {
      x0=(ix/(n1*n2));
      x1=(ix/n2)%n1;
      x2=ix%n2;

      if (my_rank==0)
      {
         n=4*n3;
         fread(vbuf,sizeof(su3_dble),n,fin);

         for (k=0;k<nc3;k++)
         {
            for (l=0;l<n;l++)
               vbuf[k*n+l]=vbuf[l];
         }
      }

      for (ic=0;ic<(nc0*nc1*nc2);ic++)
      {
         c0=(ic/(nc1*nc2));
         c1=(ic/nc2)%nc1;
         c2=ic%nc2;

         y0=x0+c0*n0;
         y1=x1+c1*n1;
         y2=x2+c2*n2;
         iy=(y2%L2)+L2*(y1%L1)+L1*L2*(y0%L0);

         np[0]=y0/L0;
         np[1]=y1/L1;
         np[2]=y2/L2;

         for (y3=0;y3<N3;y3+=L3)
         {
            np[3]=y3/L3;
            n=ipr_global(np);
            MPI_Barrier(MPI_COMM_WORLD);

            if (n>0)
            {
               tag=mpi_tag();

               if (my_rank==0)
                  MPI_Send((double*)(vbuf+4*y3),4*L3*18,MPI_DOUBLE,n,tag,
                           MPI_COMM_WORLD);

               if (my_rank==n)
                  MPI_Recv((double*)(ubuf),4*L3*18,MPI_DOUBLE,0,tag,
                           MPI_COMM_WORLD,&stat);
            }
            else if (my_rank==0)
               for (l=0;l<(4*L3);l++)
                  ubuf[l]=vbuf[4*y3+l];

            if (my_rank==n)
               set_links(iy);
         }
      }
   }

   if (my_rank==0)
   {
      error_root(ferror(fin)!=0,1,"import_cnfg [archive.c]","Read error");
      fclose(fin);
   }

   set_flags(NEW_UD);
   plaq1=plaq_sum_dble()/(double)(6*NPROC*VOLUME);
   error(fabs(plaq1-plaq0)>(DBL_EPSILON*sqrt((double)(6*NPROC*VOLUME))),1,
         "import_cnfg [archive.c]","Average plaquette is incorrect");
}
