
/*******************************************************************************
*
* File action.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the plaquette action of the global gauge fields
*
* The externally accessible functions are
*
*   double plaq_sum(void)
*     Returns the sum of Re(tr{U(p)}) over all 6*VOLUME*NPROC unoriented
*     plaquettes p, where U(p) is the product of the single-precision link
*     variables around p
*
*   double plaq_sum_dble(void)
*     Returns the sum of Re(tr{U(p)}) over all 6*VOLUME*NPROC unoriented
*     plaquettes p, where U(p) is the product of the single-precision link
*     variables around p
*
* Notes:
*
* The return values are exactly the same on all processes and the required
* communications (including buffer allocation) are performed automatically
* using the programs in the module ucom.c
*
*******************************************************************************/

#define ACTION_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "flags.h"
#include "update.h"
#include "global.h"

#define MAX_LEVELS 8
#define BLK_LENGTH 8

static int cnt[MAX_LEVELS];
static double smx[MAX_LEVELS];
static su3 w1,w2,*u[4];
static su3_dble wd1,wd2,*ud[4];


static float plaq(int ix,int mu,int nu)
{
   float sm;
   su3_vector *v1,*v2;

   plaq_u(ix,mu,nu,u);

   _su3_times_su3(w1,(*u[0]),(*u[1]));
   _su3_times_su3(w2,(*u[2]),(*u[3]));

   v1=(su3_vector*)(&w1.c11);
   v2=(su3_vector*)(&w2.c11);
   sm=_vector_prod_re((*(v1  )),(*(v2  )))+
      _vector_prod_re((*(v1+1)),(*(v2+1)))+
      _vector_prod_re((*(v1+2)),(*(v2+2)));

   return sm;
}


static double local_plaq_sum(void)
{
   int ix,mu,nu;
   double pa;

   pa=0.0;

   for (ix=0;ix<VOLUME;ix++)
   {
      for (mu=1;mu<4;mu++)
      {
         for (nu=0;nu<mu;nu++)
            pa+=(double)(plaq(ix,mu,nu));
      }
   }

   return pa;
}


double plaq_sum(void)
{
   double p,pa;

   if (query_flags(UBUF_UP2DATE)==0)
      copy_bnd_u();

   error(check_flags(ACTION),1,
         "plaq_sum [action.c]","Fields are not in proper condition");

   p=local_plaq_sum();
   MPI_Reduce(&p,&pa,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(&pa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   return pa;
}


static double plaq_dble(int ix,int mu,int nu)
{
   double sm;
   su3_vector_dble *v1,*v2;

   plaq_ud(ix,mu,nu,ud);

   _su3_times_su3(wd1,(*ud[0]),(*ud[1]));
   _su3_times_su3(wd2,(*ud[2]),(*ud[3]));

   v1=(su3_vector_dble*)(&wd1.c11);
   v2=(su3_vector_dble*)(&wd2.c11);
   sm=_vector_prod_re((*(v1  )),(*(v2  )))+
      _vector_prod_re((*(v1+1)),(*(v2+1)))+
      _vector_prod_re((*(v1+2)),(*(v2+2)));

   return sm;
}


static double local_plaq_sum_dble(void)
{
   int ix,mu,nu,n;
   double pa;

   for (n=0;n<MAX_LEVELS;n++)
   {
      cnt[n]=0;
      smx[n]=0.0;
   }
   
   for (ix=0;ix<VOLUME;ix++)
   {
      pa=0.0;

      for (mu=1;mu<4;mu++)
      {
         for (nu=0;nu<mu;nu++)
            pa+=plaq_dble(ix,mu,nu);
      }

      cnt[0]+=1;
      smx[0]+=pa;

      for (n=1;(cnt[n-1]>=BLK_LENGTH)&&(n<MAX_LEVELS);n++)
      {
         cnt[n]+=1;
         smx[n]+=smx[n-1];

         cnt[n-1]=0;
         smx[n-1]=0.0;
      }               
   }

   pa=0.0;

   for (n=0;n<MAX_LEVELS;n++)
      pa+=smx[n]; 
   
   return pa;
}


double plaq_sum_dble(void)
{
   double p,pa;

   if (query_flags(UDBUF_UP2DATE)==0)
      copy_bnd_ud();

   error(check_flags(ACTION_DBLE),1,
         "plaq_sum_dble [action.c]","Fields are not in proper condition");

   p=local_plaq_sum_dble();
   MPI_Reduce(&p,&pa,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(&pa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   return pa;
}
